# -*- coding: utf-8 -*-
"""
Connector object to establish connections between qudi modules.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['Connector']

import weakref
from typing import Any, Type, Union
from qudi.util.overload import OverloadProxy


class Connector:
    """A connector used to connect qudi modules with each other.
    """

    def __init__(
        self, interface: Union[str, Type], name: str = None, optional: bool = False
    ):
        """Initialize a Connector instance.

        Parameters
        ----------
        interface : Union[str, Type]
            Name of the interface class to connect to or the interface class itself.
        name : str, optional
            Name of the connector in qudi config. Will set attribute name if omitted.
        optional : bool, optional
            Flag indicating if the connection is mandatory (False by default).

        Raises
        ------
        AssertionError
            If `interface` is not a string or a type.
            If `name` is not `None` or a non-empty string.
            If `optional` is not a boolean.
        """
        assert isinstance(interface, (str, type)), \
            'Parameter "interface" must be an interface class or the class name as str.'
        assert name is None or (isinstance(name, str) and name), \
            'Parameter "name" must be non-empty str or None.'
        assert isinstance(optional, bool), 'Parameter "optional" must be bool type.'
        self.interface = interface if isinstance(interface, str) else interface.__name__
        self.name = name
        self.optional = optional
        self._obj_proxy = None
        self._obj_ref = lambda: None

    def __set_name__(self, owner, name):
        if self.name is None:
            self.name = name

    def __call__(self) -> Any:
        """Return reference to the module that this connector is connected to."""
        if self.is_connected:
            return self._obj_proxy
        if self.optional:
            return None
        raise RuntimeError(
            f'Connector "{self.name}" (interface "{self.interface}") is not connected.'
        )

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memodict={}):
        return self.copy()

    def __repr__(self):
        return f'{self.__module__}.Connector("{self.interface}", "{self.name}", {self.optional})'

    def __module_died_callback(self, ref=None):
        self.disconnect()

    @property
    def is_connected(self) -> bool:
        """Read-only property to check if the Connector instance is connected to a target module.

        Returns
        -------
        bool
            Connection status flag.
            - True: Connected
            - False: Disconnected
        """
        return self._obj_proxy is not None

    def connect(self, target: Any) -> None:
        """Check if target is connectible by this connector and connect.
        """
        bases = {cls.__name__ for cls in target.__class__.mro()}
        if self.interface not in bases:
            raise RuntimeError(
                f'Module "{target}" connected to connector "{self.name}" does not implement '
                f'interface "{self.interface}".'
            )
        self._obj_proxy = OverloadProxy(target, self.interface)
        self._obj_ref = weakref.ref(target, self.__module_died_callback)

    def disconnect(self) -> None:
        """Disconnect connector.
        """
        self._obj_proxy = None

    def copy(self, **kwargs):
        """Create a new instance of Connector with copied values and update
        """
        return Connector(kwargs.get('interface', self.interface),
                         kwargs.get('name', self.name),
                         kwargs.get('optional', self.optional))



# -*- coding: utf-8 -*-

"""
This file contains scripts for testing the qudi.core.scripting package.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Iterable, Sequence, Mapping, Union, Any, Optional, Tuple

from qudi.core.scripting.moduletask import ModuleTask
from qudi.core.connector import Connector


class TestTask(ModuleTask):

    _derp = Connector(name='derp', interface='TemplateLogic')

    def _setup(self) -> None:
        i = 0
        for i in range(100000000):
            i += 1

    def _cleanup(self) -> None:
        i = 0
        for i in range(100000000):
            i += 1

    def _run(self, pos_arg='abc', kw_arg=42):
        i = 0
        for i in range(10000000):
            self._check_interrupt()
            i += 1


class TestTask2(ModuleTask):

    _derp = Connector(name='derp', interface='TemplateLogic')

    def _setup(self) -> None:
        i = 0
        for i in range(100000000):
            i += 1

    def _cleanup(self) -> None:
        i = 0
        for i in range(100000000):
            i += 1

    def _run(self, seq_arg: Sequence[int], iter_arg: Iterable[str], map_arg: Mapping[str, int],
             opt_arg: Optional[int] = 42
             ) -> Tuple[Sequence[int], Iterable[str], Mapping[str, int], int]:
        i = 0
        for i in range(10000000):
            if i % 100 == 0:
                self._check_interrupt()
            i += 1
        return seq_arg, iter_arg, map_arg, opt_arg



# -*- coding: utf-8 -*-

"""
Static file handler and mixin for handling qudi configuration files.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['FileHandler', 'FileHandlerBase', 'ParserError', 'ValidationError', 'YAMLError',
           'DuplicateKeyError']


import os
from typing import Any, Dict, Mapping

from qudi.util.paths import get_default_config_dir, get_appdata_dir
from qudi.util.yaml import yaml_dump, yaml_load, ParserError, YAMLError, DuplicateKeyError

from .validator import validate_config, ValidationError


class FileHandlerBase:
    """File handler base class providing static methods for handling raw qudi configuration files.
    """

    @classmethod
    def _load(cls, path: str) -> Dict[str, Any]:
        return yaml_load(cls._relative_to_absolute_path(path))

    @classmethod
    def _dump(cls, path: str, config: Mapping[str, Any]) -> None:
        if not path.endswith('.cfg'):
            raise ValueError('Configuration file must have ".cfg" file extension.')
        path = cls._relative_to_absolute_path(path)
        os.makedirs(os.path.dirname(path), exist_ok=True)
        return yaml_dump(path, config)

    @classmethod
    def set_default_path(cls, path: str) -> None:
        """Writes the given config file path to "<AppData>/qudi/load.cfg" to be used as default
        config at the next start of qudi."""
        # Write current config file path to load.cfg
        yaml_dump(
            os.path.join(get_appdata_dir(create_missing=True), 'load.cfg'),
            {'load_config_path': cls._relative_to_absolute_path(path)}
        )

    @staticmethod
    def get_saved_path() -> str:
        """Tries to parse "<AppData>/qudi/load.cfg" and return the stored config file path.
        Raises FileNotFoundError if unsuccessful or if the recovered file path does not exist.
        """
        # Try loading config file path from last session
        load_cfg = yaml_load(os.path.join(get_appdata_dir(), 'load.cfg'), ignore_missing=True)
        file_path = load_cfg.get('load_config_path', '')
        if os.path.exists(file_path) and file_path.endswith('.cfg'):
            return file_path

        # Raise error if no last run config file could be found
        raise FileNotFoundError('No config file path saved from previous qudi sessions')

    @staticmethod
    def get_default_path() -> str:
        """Tries to find config file named "default.cfg" in several locations with the following,
        non-recursive search directory priority:
            1. <UserHome>/qudi/config/
            2. <AppData>/qudi/

        Raises FileNotFoundError if no "default.cfg" file could be found in the above locations.
        """
        # Try default.cfg in user home directory
        file_path = os.path.join(get_default_config_dir(create_missing=False), 'default.cfg')
        if os.path.exists(file_path):
            return file_path

        # Fall back to default.cfg in qudi AppData directory if possible
        file_path = os.path.join(get_appdata_dir(create_missing=False), 'default.cfg')
        if os.path.exists(file_path):
            return file_path

        # Raise error if no config file could be found
        raise FileNotFoundError('No config file could be found in default directories')

    @staticmethod
    def _relative_to_absolute_path(path):
        """Helper method converting given relative path to an existing absolute path.
        Prepends directories to given path with the following priority until an existing path has
        been created:
            1. <UserHome>/qudi/config/
            2. <AppData>/qudi/

        Raises FileNotFoundError if no existing path could be reconstructed by the above algorithm.
        """
        # absolute or relative path? Existing?
        if os.path.isabs(path) and os.path.exists(path):
            return path

        # relative path? Try relative to userdata dir, user home dir and relative to main dir
        for search_dir in [get_default_config_dir(), get_appdata_dir()]:
            new_path = os.path.abspath(os.path.join(search_dir, path))
            if os.path.exists(new_path):
                return new_path

        # Raise exception if no existing path can be determined
        raise FileNotFoundError(
            f'Qudi relative path "{path}" can not be resolved or does not exist.'
        )


class FileHandler(FileHandlerBase):
    """File handler class providing static methods for handling raw qudi configuration files.
    Also applies qudi configuration validation and default value insertion upon loading/dumping
    configuration files.
    """

    @classmethod
    def load(cls, path: str) -> Dict[str, Any]:
        """Load and validate a qudi configuration file from disk.
        Raises jsonschema.ValidationError if validation fails.
        """
        config = cls._load(path)
        validate_config(config)
        return config

    @classmethod
    def dump(cls, path: str, config: Dict[str, Any]) -> None:
        """Validate and dump a qudi configuration file to disk.
        Raises jsonschema.ValidationError if validation fails.
        """
        validate_config(config)
        cls._dump(path, config)



# -*- coding: utf-8 -*-
"""
Stand-in extension of Qt's QMutex and QRecursiveMutex classes.
Derived from the ACQ4 project.

Copyright (c) 2010, Luke Campagnola.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['Mutex', 'RecursiveMutex']

from PySide2.QtCore import QMutex as _QMutex
from PySide2.QtCore import QRecursiveMutex as _QRecursiveMutex
from typing import Optional, Union


_RealNumber = Union[int, float]


class Mutex(_QMutex):
    """Extends QMutex which serves as access serialization between threads.

    This class provides:
    * Drop-in replacement for threading.Lock
    * Context management (enter/exit)
    """

    def acquire(self, blocking: Optional[bool] = True, timeout: Optional[_RealNumber] =
-1) -> bool:
        """
        Mimics threading.Lock.acquire() to allow this class as a drop-in replacement.

        Parameters
        ----------
        blocking : bool, optional
            If True, this method will block until the mutex is locked (up to <timeout> seconds).
            If False, this method will return immediately regardless of the lock status. Default is True.
        timeout : float, optional
            Timeout in seconds specifying the maximum wait time for the mutex to be able to lock.
            Negative numbers correspond to infinite wait time. This parameter is ignored if blocking is False.
            Default is -1.0.

        """
        if blocking:
            # Convert to milliseconds for QMutex
            return self.tryLock(max(-1, int(timeout * 1000)))
        return self.tryLock()

    def release(self) -> None:
        """Mimics threading.Lock.release() to allow this class as a drop-in replacement.
        """
        self.unlock()

    def __enter__(self):
        """Enter context.

        Returns
        -------
        Mutex
            This mutex object itself.
        """
        self.lock()
        return self

    def __exit__(self, *args):
        """Exit context.

        Parameters
        ----------
        *args
            Context arguments (type, value, traceback) passed to the method.
        """
        self.unlock()


# Compatibility workaround for PySide2 vs. PySide6. In PySide2 we need to use QMutex with an
# initializer argument to construct a recursive mutex but in PySide6 we need to subclass
# QRecursiveMutex. Check if QRecursiveMutex class has all API members (indicating it's PySide6).
if all(hasattr(_QRecursiveMutex, attr) for attr in ('lock', 'unlock', 'tryLock')):
    class RecursiveMutex(_QRecursiveMutex):
        """Extends QRecursiveMutex which serves as access serialization between threads.

        This class provides:
        * Drop-in replacement for threading.Lock
        * Context management (enter/exit)

        NOTE: A recursive mutex is much more expensive than using a regular mutex. So consider
        refactoring your code to use a simple mutex before using this object.
        """

        def acquire(
            self, blocking: Optional[bool] = True, timeout: Optional[_RealNumber] = -1
        ) -> bool:
            """
            Mimics threading.Lock.acquire() to allow this class as a drop-in replacement.

            Parameters
            ----------
            blocking : bool, optional
                If True, this method will block until the mutex is locked (up to <timeout> seconds).
                If False, this method will return immediately regardless of the lock status. Default is True.
            timeout : float, optional
                Timeout in seconds specifying the maximum wait time for the mutex to be able to lock.
                Negative numbers correspond to infinite wait time. This parameter is ignored if blocking is False.
                Default is -1.0.

            Returns
            -------
            Mutex
                This mutex object itself.

            """
            if blocking:
                # Convert to milliseconds for QMutex
                return self.tryLock(max(-1, int(timeout * 1000)))
            return self.tryLock()

        def release(self) -> None:
            """Mimics threading.Lock.release() to allow this class as a drop-in replacement.
            """
            self.unlock()

        def __enter__(self):
            """
            Enter the context managed by this mutex.

            Returns
            -------
            RecursiveMutex
                This mutex object itself.

            """
            self.lock()
            return self

        def __exit__(self, *args):
            """
            Exit the context managed by this mutex.

            Parameters
            ----------
            *args
                Context arguments (type, value, traceback) passed to the method.

            """
            self.unlock()

else:
    class RecursiveMutex(Mutex):
        """Extends QRecursiveMutex which serves as access serialization between threads.

        This class provides:
        * Drop-in replacement for threading.Lock
        * Context management (enter/exit)

        NOTE: A recursive mutex is much more expensive than using a regular mutex. So consider
        refactoring your code to use a simple mutex before using this object.
        """
        def __init__(self):
            super().__init__(_QMutex.Recursive)



# -*- coding: utf-8 -*-

"""
This module provides QWidget subclasses to enter different literals (complex, dict, list, tuple,
set).

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['ComplexLineEdit', 'ComplexValidator', 'DictLineEdit', 'DictValidator', 'ListLineEdit',
           'ListValidator', 'LiteralLineEdit', 'LiteralValidator', 'SetLineEdit', 'SetValidator',
           'TupleLineEdit', 'TupleValidator']

from PySide2 import QtCore, QtGui, QtWidgets
from typing import Any, Optional, Mapping, MutableSequence, Sequence, Set, FrozenSet, Union, List
from typing import Tuple, Dict


class LiteralValidator(QtGui.QValidator):
    """
    """
    def __init__(self, parent: Optional[QtCore.QObject] = None):
        super().__init__(parent=parent)

    def validate(self, text: str, position: int) -> QtGui.QValidator.State:
        """
        """
        try:
            self.value_from_text(text)
            return self.Acceptable
        except:
            return self.Intermediate

    def fixup(self, text: str) -> str:
        return text

    def value_from_text(self, text: str) -> Any:
        return eval(text)

    def text_from_value(self, value: Any) -> str:
        return repr(value)


class ComplexValidator(QtGui.QValidator):
    """
    """
    def __init__(self, parent: Optional[QtCore.QObject] = None):
        super().__init__(parent=parent)

    def validate(self, text: str, position: int) -> QtGui.QValidator.State:
        """
        """
        try:
            self.value_from_text(text)
            return self.Acceptable
        except ValueError:
            return self.Intermediate

    def fixup(self, text: str) -> str:
        return text

    def value_from_text(self, text: str) -> complex:
        return complex(text)

    def text_from_value(self, value: complex) -> str:
        if value is None:
            value = complex()
        return repr(complex(value))


class ListValidator(QtGui.QValidator):
    """
    """
    def __init__(self, parent: Optional[QtCore.QObject] = None):
        super().__init__(parent=parent)

    def validate(self, text: str, position: int) -> QtGui.QValidator.State:
        """
        """
        try:
            self.value_from_text(text)
            return self.Acceptable
        except:
            return self.Intermediate

    def fixup(self, text: str) -> str:
        return text

    def value_from_text(self, text: str) -> List[Any]:
        tmp = eval(text)
        if isinstance(tmp, (list, tuple)):
            return list(tmp)
        raise ValueError

    def text_from_value(self, value: MutableSequence[Any]) -> str:
        if value is None:
            value = list()
        return repr(list(value))


class TupleValidator(QtGui.QValidator):
    """
    """
    def __init__(self, parent: Optional[QtCore.QObject] = None):
        super().__init__(parent=parent)

    def validate(self, text: str, position: int) -> QtGui.QValidator.State:
        """
        """
        try:
            self.value_from_text(text)
            return self.Acceptable
        except:
            return self.Intermediate

    def fixup(self, text: str) -> str:
        return text

    def value_from_text(self, text: str) -> Tuple[Any, ...]:
        tmp = eval(text)
        if isinstance(tmp, tuple):
            return tmp
        raise ValueError

    def text_from_value(self, value: Sequence[Any]) -> str:
        if value is None:
            value = tuple()
        return repr(tuple(value))


class SetValidator(QtGui.QValidator):
    """
    """
    def __init__(self, parent: Optional[QtCore.QObject] = None):
        super().__init__(parent=parent)

    def validate(self, text: str, position: int) -> QtGui.QValidator.State:
        """
        """
        try:
            self.value_from_text(text)
            return self.Acceptable
        except:
            return self.Intermediate

    def fixup(self, text: str) -> str:
        return text

    def value_from_text(self, text: str) -> Set[Any]:
        tmp = eval(text)
        if isinstance(tmp, (tuple, set, frozenset)):
            return set(tmp)
        raise ValueError

    def text_from_value(self, value: Union[Set[Any], FrozenSet[Any]]) -> str:
        if value is None:
            value = set()
        return repr(set(value))


class DictValidator(QtGui.QValidator):
    """
    """
    def __init__(self, parent: Optional[QtCore.QObject] = None):
        super().__init__(parent=parent)

    def validate(self, text: str, position: int) -> QtGui.QValidator.State:
        """
        """
        try:
            self.value_from_text(text)
            return self.Acceptable
        except:
            return self.Intermediate

    def fixup(self, text: str) -> str:
        return text

    def value_from_text(self, text: str) -> Dict[Any, Any]:
        tmp = eval(text)
        if isinstance(tmp, dict):
            return tmp
        elif isinstance(tmp, tuple):
            return dict(tmp)
        raise ValueError

    def text_from_value(self, value: Mapping[Any, Any]) -> str:
        if value is None:
            value = dict()
        return repr(dict(value))


class LiteralLineEdit(QtWidgets.QLineEdit):
    """
    """

    valueChanged = QtCore.Signal(object)

    def __init__(self, value: Optional[Any] = None, parent: Optional[QtWidgets.QWidget] = None,
                 validator: Optional[QtGui.QValidator] = None):
        super().__init__(parent=parent)
        if validator is None:
            validator = LiteralValidator()
        self._last_valid_text = ''
        self.setValidator(validator)
        self.setValue(value)

    def setValue(self, value: Any) -> None:
        """
        """
        validator = self.validator()
        text = validator.fixup(validator.text_from_value(value))
        if self._new_text_valid(text):
            self.setText(text)
            self._last_valid_text = text
            self.valueChanged.emit(self.value())
        elif text != self._last_valid_text:
            raise ValueError

    def value(self) -> Any:
        """
        """
        return self.validator().value_from_text(self._last_valid_text)

    def focusOutEvent(self, event: QtGui.QFocusEvent) -> None:
        self._revert_text()
        return super().focusOutEvent(event)

    def keyPressEvent(self, event: QtGui.QKeyEvent) -> None:
        check_text = True
        if event.key() in (QtCore.Qt.Key_Enter, QtCore.Qt.Key_Return):
            self._revert_text()
            check_text = False
        ret_val = super().keyPressEvent(event)
        if check_text:
            text = self.text()
            if self._new_text_valid(text):
                self._last_valid_text = text
                self.valueChanged.emit(self.value())
        return ret_val

    def _revert_text(self) -> None:
        text = self.text()
        if self.validator().validate(text, len(text)) != QtGui.QValidator.Acceptable:
            self.setText(self._last_valid_text)

    def _new_text_valid(self, text: str) -> bool:
        """ Helper method to check if the given text is suitable to replace the current text as
        valid value.
        """
        if text != self._last_valid_text:
            validator = self.validator()
            return validator.validate(text, len(text)) == QtGui.QValidator.Acceptable
        return False


class ComplexLineEdit(LiteralLineEdit):
    """
    """

    valueChanged = QtCore.Signal(complex)

    def __init__(self, value: Optional[complex] = None, parent: Optional[QtWidgets.QWidget] = None):
        if value is None:
            value = complex()
        super().__init__(value=value, parent=parent, validator=ComplexValidator())


class ListLineEdit(LiteralLineEdit):
    """
    """

    valueChanged = QtCore.Signal(list)

    def __init__(self, value: Optional[MutableSequence] = None,
                 parent: Optional[QtWidgets.QWidget] = None):
        if value is None:
            value = list()
        super().__init__(value=value, parent=parent, validator=ListValidator())


class TupleLineEdit(LiteralLineEdit):
    """
    """

    valueChanged = QtCore.Signal(tuple)

    def __init__(self, value: Optional[Sequence] = None,
                 parent: Optional[QtWidgets.QWidget] = None):
        if value is None:
            value = tuple()
        super().__init__(value=value, parent=parent, validator=TupleValidator())


class SetLineEdit(LiteralLineEdit):
    """
    """

    valueChanged = QtCore.Signal(set)

    def __init__(self, value: Optional[Union[Set, FrozenSet]] = None,
                 parent: Optional[QtWidgets.QWidget] = None):
        if value is None:
            value = set()
        super().__init__(value=value, parent=parent, validator=SetValidator())


class DictLineEdit(LiteralLineEdit):
    """
    """

    valueChanged = QtCore.Signal(dict)

    def __init__(self, value: Optional[Mapping] = None, parent: Optional[QtWidgets.QWidget] = None):
        if value is None:
            value = dict()
        super().__init__(value=value, parent=parent, validator=DictValidator())



# -*- coding: utf-8 -*-

"""
QWidgets serving as editors for custom configuration entries.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['CustomItemsWidget', 'CustomOptionsWidget', 'CustomConnectorsWidget']

import os
from PySide2 import QtCore, QtWidgets, QtGui
from typing import Optional, Mapping, Union, Dict, Iterable, Any

from qudi.util.paths import get_artwork_dir


class CustomItemsWidget(QtWidgets.QWidget):
    """
    """
    def __init__(self,
                 forbidden_names: Optional[Iterable[str]] = None,
                 allowed_values: Optional[Iterable[str]] = None,
                 config: Optional[Mapping[str, str]] = None,
                 parent: Optional[QtWidgets.QWidget] = None
                 ) -> None:
        super().__init__(parent=parent)

        self._forbidden_names = frozenset() if forbidden_names is None else frozenset(
            forbidden_names)
        self._allowed_values = [val.strip() for val in allowed_values] if allowed_values else None

        layout = QtWidgets.QGridLayout()
        layout.setColumnStretch(2, 1)
        self.setLayout(layout)

        icons_dir = os.path.join(get_artwork_dir(), 'icons')
        self.add_item_button = QtWidgets.QToolButton()
        self.add_item_button.setIcon(QtGui.QIcon(os.path.join(icons_dir, 'list-add')))
        self.add_item_button.clicked.connect(self._add_item_clicked)
        self.item_name_lineedit = QtWidgets.QLineEdit()
        layout.addWidget(self.add_item_button, 0, 0, 1, 1)
        layout.addWidget(self.item_name_lineedit, 0, 1, 1, 2)

        # Remove icons reused for each custom item
        self._remove_icon = QtGui.QIcon(os.path.join(icons_dir, 'list-remove'))
        # Keep track of custom item widgets
        self._item_widgets = dict()

        # Set config if given
        self.set_config(config)

    @property
    def config(self) -> Dict[str, str]:
        if self._allowed_values is None:
            config = {
                name: widgets[2].text().strip() for name, widgets in self._item_widgets.items()
            }
        else:
            config = {
                name: widgets[2].currentText() for name, widgets in self._item_widgets.items()
            }
        return config

    def set_config(self, config: Union[None, Mapping[str, str]]) -> None:
        self.clear_items()
        if config:
            for name, value in config.items():
                self.add_item(name, value)

    def add_item(self, name: str, value: Optional[str] = None) -> None:
        if not name:
            raise ValueError('Item name must be non-empty string')
        if name in self._forbidden_names:
            raise ValueError(f'Item name to add "{name}" is one of the forbidden names:\n'
                             f'{set(self._forbidden_names)}')
        if name in self._item_widgets:
            raise ValueError(f'Item name to add "{name}" is already present')

        label = QtWidgets.QLabel(f'{name}:')
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        label.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        label.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        if self._allowed_values:
            editor = QtWidgets.QComboBox()
            editor.addItem('')
            editor.addItems(self._allowed_values)
            editor.setSizeAdjustPolicy(QtWidgets.QComboBox.AdjustToContents)
        else:
            editor = QtWidgets.QLineEdit()
        remove_button = QtWidgets.QToolButton()
        remove_button.setIcon(self._remove_icon)
        remove_button.clicked.connect(lambda: self.remove_item(name))
        row = len(self._item_widgets) + 1
        layout = self.layout()
        layout.addWidget(remove_button, row, 0)
        layout.addWidget(label, row, 1)
        layout.addWidget(editor, row, 2)
        self._item_widgets[name] = (remove_button, label, editor)

    def remove_item(self, name: str) -> None:
        if name not in self._item_widgets:
            return

        layout = self.layout()

        # Remove all widgets from layout
        for button, label, editor in reversed(list(self._item_widgets.values())):
            layout.removeWidget(button)
            layout.removeWidget(label)
            layout.removeWidget(editor)

        # Delete widgets for row to remove
        button, label, editor = self._item_widgets.pop(name)
        button.clicked.disconnect()
        button.setParent(None)
        label.setParent(None)
        editor.setParent(None)
        button.deleteLater()
        label.deleteLater()
        editor.deleteLater()

        # Add all remaining widgets to layout again
        for row, (button, label, editor) in enumerate(self._item_widgets.values(), 1):
            layout.addWidget(button, row, 0)
            layout.addWidget(label, row, 1)
            layout.addWidget(editor, row, 2)

    def clear_items(self) -> None:
        layout = self.layout()
        widgets = list(self._item_widgets.values())
        self._item_widgets.clear()
        for button, label, editor in reversed(widgets):
            layout.removeWidget(button)
            layout.removeWidget(label)
            layout.removeWidget(editor)
            button.setParent(None)
            label.setParent(None)
            editor.setParent(None)
            button.deleteLater()
            label.deleteLater()
            editor.deleteLater()

    @QtCore.Slot()
    def _add_item_clicked(self) -> None:
        name = self.item_name_lineedit.text().strip()
        try:
            self.add_item(name)
            self.item_name_lineedit.clear()
        except ValueError:
            pass


class CustomOptionsWidget(CustomItemsWidget):
    """
    """
    def __init__(self,
                 forbidden_names: Optional[Iterable[str]] = None,
                 config: Optional[Mapping[str, Any]] = None,
                 parent: Optional[QtWidgets.QWidget] = None
                 ) -> None:
        super().__init__(forbidden_names=forbidden_names, config=config, parent=parent)

        self.add_item_button.setToolTip('Add custom ConfigOption with given name.')
        self.item_name_lineedit.setPlaceholderText('Enter custom ConfigOption name')

    @property
    def config(self) -> Dict[str, Any]:
        cfg = super().config
        for name, value in cfg.items():
            if value == '':
                cfg[name] = None
            else:
                try:
                    cfg[name] = eval(value)
                except:
                    pass
        return cfg

    def set_config(self, config: Union[None, Mapping[str, Any]]) -> None:
        if config:
            config = {name: '' if val is None else repr(val) for name, val in config.items()}
        return super().set_config(config)

    def add_item(self, name: str, value: Optional[str] = None) -> None:
        super().add_item(name=name, value=value)
        self._item_widgets[name][2].setPlaceholderText('text parsed by eval()')


class CustomConnectorsWidget(CustomItemsWidget):
    """
    """
    def __init__(self,
                 forbidden_names: Optional[Iterable[str]] = None,
                 module_names: Optional[Iterable[str]] = None,
                 config: Optional[Mapping[str, Any]] = None,
                 parent: Optional[QtWidgets.QWidget] = None
                 ) -> None:
        super().__init__(forbidden_names=forbidden_names,
                         allowed_values=module_names,
                         config=config,
                         parent=parent)

        self.add_item_button.setToolTip('Add custom Connector with given name.')
        self.item_name_lineedit.setPlaceholderText('Enter custom Connector name')



# -*- coding: utf-8 -*-

"""
This file contains models for linear fitting routines for qudi based on the lmfit package.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ('Linear',)

import numpy as np
from qudi.util.fit_models.model import FitModelBase, estimator


class Linear(FitModelBase):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_param_hint('offset', value=0, min=-np.inf, max=np.inf)
        self.set_param_hint('slope', value=0, min=-np.inf, max=np.inf)

    @staticmethod
    def _model_function(x, offset, slope):
        return offset + slope * x

    @estimator('default')
    def estimate(self, data, x):
        data = np.asarray(data)
        x = np.asarray(x)
        # calculate the parameters using Least-squares estimation of linear regression
        x_mean = np.mean(x)
        data_mean = np.mean(data)
        a_1 = np.sum((x - x_mean) * (data - data_mean))
        a_2 = np.sum(np.power(x - x_mean, 2))

        slope = a_1 / a_2
        intercept = data_mean - slope * x_mean

        max_slope = (max(data) - min(data)) / abs(x[-1] - x[0])  # maximum slope possible

        estimate = self.make_params()
        estimate['offset'].set(value=intercept, min=-np.inf, max=np.inf)
        estimate['slope'].set(value=slope, min=-max_slope, max=max_slope)
        return estimate

    @estimator('No Offset')
    def estimate_no_offset(self, data, x):
        estimate = self.estimate(data, x)
        estimate['offset'].set(value=0, min=-np.inf, max=np.inf, vary=False)
        return estimate

    @estimator('Constant')
    def estimate_no_offset(self, data, x):
        estimate = self.make_params()
        estimate['slope'].set(value=0, min=-np.inf, max=np.inf, vary=False)
        estimate['offset'].set(value=np.mean(data), min=min(data), max=max(data))
        return estimate



# -*- coding: utf-8 -*-

"""
Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['ModuleSelector']

from PySide2 import QtWidgets, QtCore
from typing import Optional, Iterable, List, Mapping, Tuple, Dict

from qudi.tools.config_editor.tree_widgets import AvailableModulesTreeWidget
from qudi.tools.config_editor.tree_widgets import SelectedModulesTreeWidget
from qudi.util.widgets.separator_lines import HorizontalLine


class ModuleSelector(QtWidgets.QDialog):
    """QDialog representing a selection editor for qudi modules to configure.
    """

    def __init__(self,
                 available_modules: Iterable[str],
                 named_modules: Optional[Mapping[str, str]] = None,
                 unnamed_modules: Optional[Iterable[str]] = None,
                 parent: Optional[QtWidgets.QWidget] = None
                 ) -> None:
        super().__init__(parent=parent)

        self.setWindowTitle('Qudi Config Editor: Module Selection')
        screen_size = QtWidgets.QApplication.instance().primaryScreen().availableSize()
        self.resize(screen_size.width() // 3, screen_size.height() // 3)

        # Create two customized QTreeWidgets. One for all available modules to select from and one
        # for the selected modules.
        self.available_treewidget = AvailableModulesTreeWidget(modules=available_modules)
        self.selected_treewidget = SelectedModulesTreeWidget(named_modules=named_modules,
                                                             unnamed_modules=unnamed_modules)

        # Create left side of splitter widget
        left_widget = QtWidgets.QWidget()
        layout = QtWidgets.QVBoxLayout()
        label = QtWidgets.QLabel('Available Modules')
        font = label.font()
        font.setPointSize(font.pointSize() + 4)
        font.setBold(True)
        label.setFont(font)
        layout.addWidget(label)
        layout.addWidget(self.available_treewidget)
        left_widget.setLayout(layout)

        # Create right side of splitter widget
        right_widget = QtWidgets.QWidget()
        layout = QtWidgets.QGridLayout()
        right_widget.setLayout(layout)
        label = QtWidgets.QLabel('Selected Modules')
        label.setFont(font)
        self.add_remote_button = QtWidgets.QPushButton('Add Remote Module')
        self.add_custom_button = QtWidgets.QPushButton('Add Custom Module')
        self.custom_module_lineedit = QtWidgets.QLineEdit()
        self.custom_module_lineedit.setPlaceholderText('Custom module name (module.Class)')
        self.base_selection_combobox = QtWidgets.QComboBox()
        self.base_selection_combobox.addItems(['GUI', 'Logic', 'Hardware'])
        self.add_remote_button.clicked.connect(self._add_remote_module_clicked)
        self.add_custom_button.clicked.connect(self._add_custom_module_clicked)
        layout.addWidget(label, 0, 0, 1, 3)
        layout.addWidget(self.selected_treewidget, 1, 0, 1, 3)
        label = QtWidgets.QLabel('Module Base:')
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        layout.addWidget(label, 2, 0)
        layout.addWidget(self.base_selection_combobox, 2, 1)
        layout.addWidget(self.add_remote_button, 2, 2)
        layout.addWidget(self.custom_module_lineedit, 3, 0, 1, 2)
        layout.addWidget(self.add_custom_button, 3, 2)
        layout.setColumnStretch(0, 1)
        layout.setRowStretch(1, 1)

        # set splitter as main widget
        splitter = QtWidgets.QSplitter()
        splitter.addWidget(left_widget)
        splitter.addWidget(right_widget)
        splitter.setCollapsible(0, False)
        splitter.setCollapsible(1, False)

        # Create buttonbox for this dialog

        label = QtWidgets.QLabel(
            'Include qudi modules by dragging them into the right field (press DEL to remove).'
        )
        label.setAlignment(QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter)
        font = label.font()
        font.setBold(True)
        label.setFont(font)
        self.button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Cancel | QtWidgets.QDialogButtonBox.Ok
        )
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)
        sub_layout = QtWidgets.QHBoxLayout()
        sub_layout.addWidget(label)
        sub_layout.addStretch()
        sub_layout.addWidget(self.button_box)

        # Add everything to the main layout
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(splitter)
        layout.addWidget(HorizontalLine())
        layout.addLayout(sub_layout)
        layout.setStretch(0, 1)
        self.setLayout(layout)

    @property
    def selected_modules(self) -> Tuple[Dict[str, str], List[str]]:
        return self.selected_treewidget.modules

    @QtCore.Slot()
    def _add_remote_module_clicked(self) -> None:
        base = self.base_selection_combobox.currentText().lower()
        if base == 'gui':
            raise ValueError('Unable to add remote module.\nGUI modules can not be remote modules.')
        self.selected_treewidget.add_module(f'{base}.<REMOTE MODULE>')

    @QtCore.Slot()
    def _add_custom_module_clicked(self) -> None:
        base = self.base_selection_combobox.currentText().lower()
        module = self.custom_module_lineedit.text().strip()
        if module:
            self.selected_treewidget.add_module(f'{base}.{module}')



# -*- coding: utf-8 -*-

"""
This file contains custom QWidgets to show (animated) loading indicators.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

from PySide2 import QtWidgets, QtCore, QtGui


class CircleLoadingIndicator(QtWidgets.QWidget):
    """ Simple circular loading indicator.
    You can customize cycle period, indicator arc length and width.
    Animation will automatically start (stop) upon showing (hiding) the widget.
    The widget can be arbitrarily resized but the actual indicator will always maintain 1:1 aspect
    ratio and will be centered.
    The color of the indicator is chosen to be the current palette highlight color.

    Indicator length must be specified as integer value in 1/16th of a degree.
    Indicator width ratio can be any value 0 < x <= 0.5
    """

    def __init__(self, *args, cycle_time=1.2, indicator_length=960, indicator_width_ratio=0.2,
                 **kwargs):
        """
        Parameters
        ----------
        cycle_time : float
            The animation time in seconds for a full cycle.
        indicator_length : int
            Length of the indicator arc in 1/16th of a degree.
        indicator_width_ratio : float
            Ratio of the indicator arc width with respect to the widget size.

        Returns
        -------
        None
            (or specify the return type and description if the function returns something)
        """
        assert cycle_time > 0, 'cycle_time must be larger than 0'
        assert 0 < indicator_length < 5760, 'indicator_length must be >0 and <5760'
        assert 0 < indicator_width_ratio <= 0.5, 'indicator_width_ratio must be >0 and <=0.5'
        super().__init__(*args, **kwargs)
        self.setMinimumSize(6, 6)
        self.setMouseTracking(False)
        self.setFocusPolicy(QtCore.Qt.NoFocus)

        # Fixed init parameters
        self._indicator_length = indicator_length
        self._cycle_time_ms = int(round(1000 * cycle_time))
        self._indicator_width_ratio = indicator_width_ratio

        # property value (angle in 1/16th of a degree) for current indicator position.
        # 0 means 3 o'clock.
        self._indicator_position = 0

        # misc parameters
        self.__animation = None
        self.__pen = QtGui.QPen(self.palette().highlight().color())
        self.__pen.setCapStyle(QtCore.Qt.RoundCap)
        self.__draw_rect = None
        self.__update_draw_size()
        self.__size_hint = None
        self.__update_size_hint()

    @QtCore.Property(int)
    def indicator_position(self):
        return self._indicator_position

    @indicator_position.setter
    def indicator_position(self, value):
        self._indicator_position = value
        self.update()

    def sizeHint(self):
        return self.__size_hint

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self.__update_draw_size()
        self.__update_size_hint()

    def paintEvent(self, event):
        # Set up painter
        p = QtGui.QPainter(self)
        p.setRenderHint(QtGui.QPainter.Antialiasing, True)
        p.setBrush(QtCore.Qt.NoBrush)
        self.__pen.setColor(self.palette().highlight().color())  # in case the palette has changed
        p.setPen(self.__pen)

        # draw indicator
        p.drawArc(self.__draw_rect, self._indicator_position, self._indicator_length)

    def showEvent(self, ev):
        super().showEvent(ev)
        if self.__animation is None:
            self.__animation = QtCore.QPropertyAnimation(self, b'indicator_position', self)
            self.__animation.setDuration(self._cycle_time_ms)
            self.__animation.setStartValue(0)
            self.__animation.setEndValue(-5760)
            self.__animation.setLoopCount(-1)
            self.__animation.start()

    def hideEvent(self, ev):
        if self.__animation is not None:
            self.__animation.stop()
            self.__animation = None
        super().hideEvent(ev)

    def __update_draw_size(self):
        width = self.width()
        height = self.height()
        if height > width:
            x_offset = 0
            y_offset = (height - width) // 2
            base_size = width
        else:
            x_offset = (width - height) // 2
            y_offset = 0
            base_size = height
        line_width = max(1, int(round(base_size * self._indicator_width_ratio)))
        margin = max(1, line_width // 2)
        size = base_size - 2 * margin
        self.__draw_rect = QtCore.QRect(x_offset + margin, y_offset + margin, size, size)
        self.__pen.setWidth(line_width)

    def __update_size_hint(self):
        self.__size_hint = QtCore.QSize(min(self.width(), self.height()),
                                        min(self.width(), self.height()))



# -*- coding: utf-8 -*-
"""
Decorators and objects used for qudi interfaces

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ('ScalarConstraint',)

import warnings
warnings.warn('The "qudi.core.interface" module is deprecated and will soon be removed in future '
              'releases. Please use the objects contained in "qudi.util.constraints" instead.',
              DeprecationWarning,
              stacklevel=2)


class ScalarConstraint:
    """
    Constraint definition for a scalar variable hardware parameter.
    """
    def __init__(self, min=0.0, max=0.0, step=0.0, default=0.0, unit=''):
        # allowed minimum value for parameter
        self.min = min
        # allowed maximum value for parameter
        self.max = max
        # allowed step size for parameter value changes (for spinboxes etc.)
        self.step = step
        # the default value for the parameter
        self.default = default
        # the unit of the parameter value(optional)
        self.unit = unit
        return



# -*- coding: utf-8 -*-
"""
ConfigOption object to be used in qudi modules. The value of each ConfigOption can
(if it has a default value) or must be specified by the user in the config file.
Usually these values should be constant for the duration of a qudi session.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['ConfigOption', 'MissingOption']

import copy
import inspect
from enum import Enum
from typing import Any, Optional, Callable


class MissingOption(Enum):
    """ Representation for missing ConfigOption """
    error = -3
    warn = -2
    info = -1
    nothing = 0


class ConfigOption:
    """This class represents a configuration entry in the config file that is loaded before
    module initalisation.
    """

    def __init__(self, name: Optional[str] = None, default: Optional[Any] = None, *,
                 missing: Optional[str] = 'nothing', constructor: Optional[Callable] = None,
                 checker: Optional[Callable] = None, converter: Optional[Callable] = None):
        """ Create a ConfigOption object.

        Parameters
        ----------
        name : Optional[str], optional
            Identifier of the option in the configuration file.
        default : Optional[Any], optional
            Default value for the case that the option is not set in the config file.
        missing : Optional[str], optional
            Action to take when the option is not set.
            'nothing' does nothing,
            'warn' logs a warning,
            'error' logs an error and prevents the module from loading.
        constructor : Optional[Callable], optional
            Constructor function for complex config option behavior.
        checker : Optional[Callable], optional
            Static function that checks if value is okay.
        converter : Optional[Callable], optional
            Static function that forces type interpretation.

        """
        self.missing = MissingOption[missing]

        self.name = name
        self.default = default
        self.checker = checker
        self.converter = converter
        self.constructor_function = None
        if constructor is not None:
            self.constructor(constructor)

    def __set_name__(self, owner, name):
        if self.name is None:
            self.name = name

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memodict={}):
        return self.copy()

    @property
    def optional(self) -> bool:
        return self.missing != MissingOption.error

    def copy(self, **kwargs):
        """Create a new instance of ConfigOption with copied values and update.

        Parameters
        ----------
        **kwargs
            Extra arguments or overrides for the constructor of this class.
        """
        newargs = {'name': self.name,
                   'default': copy.deepcopy(self.default),
                   'missing': self.missing.name,
                   'constructor': self.constructor_function,
                   'checker': self.checker,
                   'converter': self.converter}
        newargs.update(kwargs)
        return ConfigOption(**newargs)

    def check(self, value: Any) -> bool:
        """If checker function set, check value. Assume everything is ok otherwise.
        """
        if callable(self.checker):
            return self.checker(value)
        return True

    def convert(self, value: Any) -> Any:
        """If converter function set, convert value (pass-through otherwise).
        """
        if callable(self.converter):
            return self.converter(value)
        return value

    def constructor(self, func: Callable) -> Callable:
        """Decorator for declaring a constructor function for this ConfigOption.

        Parameters
        ----------
        func : Callable
            Constructor function for this ConfigOption.

        Returns
        -------
        Callable
            Returns the original function so it can be used as a decorator.
        """
        self.constructor_function = self._assert_func_signature(func)
        return func

    @staticmethod
    def _assert_func_signature(func: Callable) -> Callable:
        assert callable(func), 'ConfigOption constructor must be callable'
        params = tuple(inspect.signature(func).parameters)
        assert 0 < len(params) < 3, 'ConfigOption constructor must be function with ' \
                                    '1 (static) or 2 (bound method) parameters.'
        if len(params) == 1:
            def wrapper(instance, value):
                return func(value)

            return wrapper
        return func



# -*- coding: utf-8 -*-

"""
Configuration editor App for creation and editing of qudi configuration files.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ('main', 'ConfigurationEditorMainWindow', 'ConfigurationEditor')

import os
import sys
from typing import Optional, Mapping, Dict, Any
from PySide2 import QtCore, QtGui, QtWidgets
from qudi.util.paths import get_main_dir, get_default_config_dir, get_artwork_dir
from qudi.core.config import Configuration

from qudi.tools.config_editor.module_selector import ModuleSelector
from qudi.tools.config_editor.module_editor import ModuleEditorWidget
from qudi.tools.config_editor.global_editor import GlobalEditorWidget
from qudi.tools.config_editor.tree_widgets import ConfigModulesTreeWidget
from qudi.tools.config_editor.module_finder import QudiModules

try:
    import matplotlib
    matplotlib.use('agg')
except ImportError:
    pass

# Enable the High DPI scaling support of Qt5
os.environ['QT_ENABLE_HIGHDPI_SCALING'] = '1'

if sys.platform == 'win32':
    # Set QT_LOGGING_RULES environment variable to suppress qt.svg related warnings that otherwise
    # spam the log due to some known Qt5 bugs, e.g. https://bugreports.qt.io/browse/QTBUG-52079
    os.environ['QT_LOGGING_RULES'] = 'qt.svg.warning=false'
else:
    # The following will prevent Qt to spam the logs on X11 systems with enough messages
    # to significantly slow the program down. Most of those warnings should have been
    # notice level or lower. This is a known problem since Qt does not fully comply to X11.
    os.environ['QT_LOGGING_RULES'] = '*.debug=false;*.info=false;*.notice=false;*.warning=false'


class ConfigurationEditor(QtWidgets.QMainWindow):
    """
    """
    def __init__(self, parent: Optional[QtWidgets.QWidget] = None) -> None:
        super().__init__(parent=parent)
        self.setWindowTitle('Qudi Config Editor')
        screen_size = QtWidgets.QApplication.instance().primaryScreen().availableSize()
        self.resize((screen_size.width() * 3) // 4, (screen_size.height() * 3) // 4)

        self.qudi_environment = QudiModules()

        self.module_tree_widget = ConfigModulesTreeWidget()
        self.module_tree_widget.itemChanged.connect(self._module_renamed_by_tree)
        self.module_tree_widget.itemSelectionChanged.connect(self._module_selection_changed)

        self.module_config_editor = ModuleEditorWidget(qudi_modules=self.qudi_environment)
        self.module_config_editor.sigModuleRenamed.connect(self._module_renamed_by_editor)

        self.global_config_editor = GlobalEditorWidget()

        label = QtWidgets.QLabel('Included Modules')
        label.setAlignment(QtCore.Qt.AlignCenter)
        font = label.font()
        font.setBold(True)
        font.setPointSize(font.pointSize() + 4)
        label.setFont(font)
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(label)
        layout.addWidget(self.module_tree_widget)
        layout.addWidget(self.global_config_editor)
        layout.setStretch(1, 1)
        left_widget = QtWidgets.QWidget()
        left_widget.setLayout(layout)

        label = QtWidgets.QLabel('Module Configuration')
        label.setAlignment(QtCore.Qt.AlignCenter)
        label.setFont(font)
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(label)
        layout.addWidget(self.module_config_editor)
        label = QtWidgets.QLabel('Mandatory fields/options/connectors are marked with *')
        font = label.font()
        font.setBold(True)
        label.setFont(font)
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        layout.addWidget(label)
        layout.setStretch(1, 1)
        right_widget = QtWidgets.QWidget()
        right_widget.setLayout(layout)

        splitter = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
        splitter.addWidget(left_widget)
        splitter.addWidget(right_widget)
        splitter.setChildrenCollapsible(False)
        splitter.setStretchFactor(0, 2)
        splitter.setStretchFactor(1, 3)
        self.setCentralWidget(splitter)

        # Main window actions
        icon_dir = os.path.join(get_main_dir(), 'artwork', 'icons')
        quit_icon = QtGui.QIcon(os.path.join(icon_dir, 'application-exit'))
        self.quit_action = QtWidgets.QAction(quit_icon, 'Quit')
        self.quit_action.setShortcut(QtGui.QKeySequence('Ctrl+Q'))
        load_icon = QtGui.QIcon(os.path.join(icon_dir, 'document-open'))
        self.load_action = QtWidgets.QAction(load_icon, 'Load')
        self.load_action.setShortcut(QtGui.QKeySequence('Ctrl+L'))
        self.load_action.setToolTip('Load a qudi configuration to edit from file.')
        save_icon = QtGui.QIcon(os.path.join(icon_dir, 'document-save'))
        self.save_action = QtWidgets.QAction(save_icon, 'Save')
        self.save_action.setShortcut(QtGui.QKeySequence('Ctrl+S'))
        self.save_action.setToolTip('Save the current qudi configuration to file.')
        self.save_as_action = QtWidgets.QAction('Save as ...')
        new_icon = QtGui.QIcon(os.path.join(icon_dir, 'document-new'))
        self.new_action = QtWidgets.QAction(new_icon, 'New')
        self.new_action.setShortcut(QtGui.QKeySequence('Ctrl+N'))
        self.new_action.setToolTip('Create a new qudi configuration from scratch.')
        select_icon = QtGui.QIcon(os.path.join(icon_dir, 'configure'))
        self.select_modules_action = QtWidgets.QAction(select_icon, 'Select Modules')
        self.select_modules_action.setShortcut(QtGui.QKeySequence('Ctrl+M'))
        self.select_modules_action.setToolTip(
            'Open an editor to select the modules to include in config.'
        )
        # Connect actions
        self.quit_action.triggered.connect(self.close)
        self.new_action.triggered.connect(self.new_config)
        self.load_action.triggered.connect(self.prompt_load_config)
        self.save_action.triggered.connect(self.save_config)
        self.save_as_action.triggered.connect(self.prompt_save_config)
        self.select_modules_action.triggered.connect(self.select_modules)

        # Create menu bar
        menu_bar = QtWidgets.QMenuBar()
        file_menu = QtWidgets.QMenu('File')
        menu_bar.addMenu(file_menu)
        file_menu.addAction(self.new_action)
        file_menu.addSeparator()
        file_menu.addAction(self.load_action)
        file_menu.addAction(self.save_action)
        file_menu.addAction(self.save_as_action)
        file_menu.addSeparator()
        file_menu.addAction(self.quit_action)
        file_menu = QtWidgets.QMenu('Edit')
        menu_bar.addMenu(file_menu)
        file_menu.addAction(self.select_modules_action)
        self.setMenuBar(menu_bar)

        # Create toolbar
        toolbar = QtWidgets.QToolBar()
        toolbar.setToolButtonStyle(QtCore.Qt.ToolButtonTextUnderIcon)
        toolbar.addAction(self.new_action)
        toolbar.addAction(self.load_action)
        toolbar.addAction(self.save_action)
        toolbar.addSeparator()
        toolbar.addAction(self.select_modules_action)
        toolbar.setFloatable(False)
        self.addToolBar(toolbar)

        # Process variables
        self._editor_item = None
        self._config_map = dict()
        self._current_file_path = None

    @QtCore.Slot()
    def _module_selection_changed(self):
        selected_items = self.module_tree_widget.selectedItems()
        if not selected_items:
            self.module_config_editor.close_editor()
            self._editor_item = None
            return

        item = selected_items[0]
        if item is not self._editor_item:
            base = item.parent().text(0).lower()
            module = f'{base}.{item.text(2)}'
            name = item.text(1)

            # Get current module config dict
            config = self._config_map.get(base, dict()).get(name, None)

            # Sort out available connectors and targets as well as module config options
            if module == '<REMOTE MODULE>':
                self.module_config_editor.open_remote_module(name, config=config)
            else:
                self.module_config_editor.open_local_module(
                    module_class=module,
                    named_modules=self.module_tree_widget.modules[0],
                    name=name,
                    config=config
                )
            self._editor_item = item

    @QtCore.Slot()
    def select_modules(self):
        self.module_config_editor.close_editor()
        available = self.qudi_environment.available_modules
        named_selected, unnamed_selected = self.module_tree_widget.modules
        selector_dialog = ModuleSelector(available_modules=available,
                                         named_modules=named_selected,
                                         unnamed_modules=unnamed_selected)
        if selector_dialog.exec_():
            # Recycle old module names if identical modules are selected but not named
            new_named_selected, new_unnamed_selected = selector_dialog.selected_modules
            recycled_named_selected = {
                name: mod for name, mod in named_selected.items() if
                (name not in new_named_selected) and (mod in new_unnamed_selected)
            }
            new_named_selected.update(recycled_named_selected)
            for mod in recycled_named_selected.values():
                new_unnamed_selected.remove(mod)
            # Set modules in main window
            self.module_tree_widget.set_modules(named_modules=new_named_selected,
                                                unnamed_modules=new_unnamed_selected)

    def new_config(self):
        self._current_file_path = None
        self._config_map = dict()
        self.module_config_editor.close_editor()
        self.module_tree_widget.set_modules(dict())
        self.global_config_editor.open_editor(None)
        self.select_modules()

    def prompt_load_config(self):
        file_path = QtWidgets.QFileDialog.getOpenFileName(
            self,
            'Qudi Config Editor: Load Configuration...',
            get_default_config_dir(),
            'Config files (*.cfg)')[0]
        if file_path:
            self.module_config_editor.close_editor()
            config = Configuration()
            config.load(file_path, set_default=False)
            self._config_map = config.config_map
            self._current_file_path = file_path
            modules = self._get_modules_from_config(self._config_map)
            global_cfg = self._config_map.get('global', dict())
            self.module_tree_widget.set_modules(named_modules=modules)
            self.global_config_editor.open_editor(global_cfg)

    def prompt_save_config(self):
        file_path = QtWidgets.QFileDialog.getSaveFileName(
            self,
            'Qudi Config Editor: Save Configuration...',
            get_default_config_dir() if self._current_file_path is None else os.path.dirname(
                self._current_file_path),
            'Config files (*.cfg)')[0]
        if file_path:
            config = Configuration(config=self._config_map)
            config.dump(file_path)
            self._current_file_path = file_path

    def prompt_overwrite(self, file_path):
        answer = QtWidgets.QMessageBox.question(
            self,
            'Qudi Config Editor: Overwrite?',
            f'Do you really want to overwrite existing Qudi configuration at\n"{file_path}"?',
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
            QtWidgets.QMessageBox.No)
        return answer == QtWidgets.QMessageBox.Yes

    def save_config(self):
        if self._current_file_path is None:
            self.prompt_save_config()
        elif os.path.exists(self._current_file_path):
            if self.prompt_overwrite(self._current_file_path):
                config = Configuration(config=self._config_map)
                config.dump(self._current_file_path)
            else:
                self.prompt_save_config()
        else:
            config = Configuration(config=self._config_map)
            config.dump(self._current_file_path)

    def prompt_close(self):
        answer = QtWidgets.QMessageBox.question(
            self,
            'Qudi Config Editor: Quit?',
            'Do you really want to quit the Qudi configuration editor?\nAll unsaved work will be '
            'lost.',
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
            QtWidgets.QMessageBox.No
        )
        return answer == QtWidgets.QMessageBox.Yes

    def closeEvent(self, event):
        if self.prompt_close():
            event.accept()
        else:
            event.ignore()

    @staticmethod
    def _get_modules_from_config(config: Mapping[str, Any]) -> Dict[str, str]:
        modules = {name: 'gui.' + cfg.get('module.Class', '<REMOTE MODULE>') for name, cfg in
                   config.get('gui', dict()).items()}
        modules.update({name: 'logic.' + cfg.get('module.Class', '<REMOTE MODULE>') for name, cfg in
                        config.get('logic', dict()).items()})
        modules.update(
            {name: 'hardware.' + cfg.get('module.Class', '<REMOTE MODULE>') for name, cfg in
             config.get('hardware', dict()).items()}
        )
        return modules

    @QtCore.Slot(QtWidgets.QTreeWidgetItem, int)
    def _module_renamed_by_tree(self, item, column):
        if column != 1 or item is None or item.parent() is None:
            return

        name = item.text(1)
        if item is self._editor_item:
            self.module_config_editor.set_module_name(name)

    @QtCore.Slot(str)
    def _module_renamed_by_editor(self, name: str) -> None:
        self.module_tree_widget.blockSignals(True)
        try:
            self._editor_item.setText(1, name)
        except AttributeError:
            pass
        finally:
            self.module_tree_widget.blockSignals(False)

    def keyPressEvent(self, event: QtGui.QKeyEvent) -> None:
        if event.key() == QtCore.Qt.Key_Escape:
            self.module_tree_widget.clearSelection()
            event.accept()
        else:
            super().keyPressEvent(event)


class ConfigurationEditorApp(QtWidgets.QApplication):
    """
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        app_icon = QtGui.QIcon(os.path.join(get_artwork_dir(), 'logo', 'logo-qudi.svg'))
        self.setWindowIcon(app_icon)


def main():
    app = ConfigurationEditorApp(sys.argv)
    # init editor QMainWindow and show
    editor = ConfigurationEditor()
    editor.show()
    # Start event loop
    sys.exit(app.exec_())


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""
This file contains helper methods to find and estimate multiple peaks/dips for fit models.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ('correct_offset_histogram', 'find_highest_peaks', 'estimate_double_peaks',
           'estimate_triple_peaks', 'sort_check_data', 'smooth_data')

import numpy as np
from scipy.signal import find_peaks as _find_peaks
from scipy.signal import peak_widths as _peak_widths
from scipy.ndimage.filters import gaussian_filter1d as _gaussian_filter


def sort_check_data(data, x):
    # sort x in increasing order and data accordingly
    sorted_args = np.argsort(x)
    x = np.asarray(x[sorted_args])
    data = np.asarray(data[sorted_args])
    return data, x


def smooth_data(data, filter_width=None):
    if filter_width is None:
        filter_width = max(1, int(round(len(data) / 100)))
    return _gaussian_filter(data, sigma=filter_width), filter_width


def correct_offset_histogram(data, bin_width=None):
    """ Subtracts a constant offset from a copy of given data array and returns it.
    The offset is assumed to be the most common value in data. This value is determined by creating
    a histogram of <data> with bin width <bin_width> and taking the value with the most occurrences.
    If no bin width has been specified, assume bin width of 1/50th of data length (min. 1).

    For best results, make sure to filter noisy data beforehand. The used smoothing filter width
    is a good estimate for optimal bin_width.

    Parameters
    ----------
    data : iterable
        Peak data to correct offset for. Must be convertible using numpy.asarray.
    bin_width : int, optional
        Bin width in samples to use for histogram creation. Default is None, which sets bin width
        to 1/50th of data length (minimum 1).

    Returns
    -------
    numpy.ndarray, float
        New array with offset-corrected data, the offset value.

    """
    data = np.asarray(data)
    if bin_width is None:
        bin_width = max(1, data.size // 50)
    elif not isinstance(bin_width, int):
        bin_width = max(1, int(round(bin_width)))
    hist = np.histogram(data, bins=bin_width)
    offset = (hist[1][hist[0].argmax()] + hist[1][hist[0].argmax() + 1]) / 2
    return data - offset, offset


def find_highest_peaks(data, peak_count, allow_borders=True, **kwargs):
    """ Find peaks using scipy.signal.find_peaks().
    ToDo: Document
    """
    peak_count = int(peak_count)
    assert peak_count > 0, 'Parameter "peak_count" must be integer >= 1'
    assert len(data) >= 5, 'Data must contain at least 5 data points'

    # Return early if all elements are the same
    if min(data) == max(data):
        return list(), list(), list()

    # Find all peaks
    peaks, properties = _find_peaks(data, **kwargs)
    if len(peaks) == 0:
        # ToDo: warn
        return list(), list(), list()

    # Sort found peaks by increasing peak height
    sorted_args = np.argsort(data[peaks])
    peaks = peaks[sorted_args]

    # Only keep requested number of highest peaks
    peaks = peaks[-peak_count:]
    peak_heights = data[peaks]
    peak_widths = _peak_widths(data, peaks, rel_height=0.5)[0]  # full-width at half-maximum

    # Check if data borders are more promising as peak locations and replace found peaks
    if allow_borders:
        width = max(2, int(round(max(peak_widths))))
        left_mean = np.mean(data[:width])
        right_mean = np.mean(data[-width:])
        if 2 * min(peak_heights) < left_mean and min(peaks) > 2 * width:
            min_arg = np.argmin(peak_heights)
            peaks[min_arg] = np.argmax(data[:width])
            peak_heights[min_arg] = data[peaks[min_arg]]
        if 2 * min(peak_heights) < 2 * right_mean and max(peaks) < len(data) - 1 - 2 * width:
            min_arg = np.argmin(peak_heights)
            peaks[min_arg] = np.argmax(data[-width:])
            peak_heights[min_arg] = data[peaks[min_arg]]
        # Check if some peaks are missing and manually add borders if they look promising
        if len(peaks) < peak_count:
            threshold = max(data) / 2
            no_left_peak = min(peaks) > 2 * width
            no_right_peak = max(peaks) < len(data) - 1 - 2 * width
            if no_left_peak and left_mean > threshold:
                peaks = np.append(peaks, np.argmax(data[:width]))
                peak_heights = np.append(peak_heights, data[peaks[-1]])
                peak_widths = np.append(peak_widths, width)
            if no_right_peak and right_mean > threshold:
                peaks = np.append(peaks, np.argmax(data[-width:]))
                peak_heights = np.append(peak_heights, data[peaks[-1]])
                peak_widths = np.append(peak_widths, width)

    return peaks, peak_heights, peak_widths


def estimate_double_peaks(data, x, minimum_distance=None):
    # Find peaks along with width and amplitude estimation
    peak_indices, peak_heights, peak_widths = find_highest_peaks(data,
                                                                 peak_count=2,
                                                                 width=minimum_distance,
                                                                 height=0.05 * max(data))

    x_spacing = min(abs(np.ediff1d(x)))
    x_span = abs(x[-1] - x[0])
    data_span = abs(max(data) - min(data))

    # Replace missing peaks with sensible default value
    if len(peak_indices) == 1:
        # If just one peak was found, assume it is two peaks overlapping and split it into two
        left_peak_index = max(0, int(round(peak_indices[0] - peak_widths[0] / 2)))
        right_peak_index = min(len(x) - 1, int(round(peak_indices[0] + peak_widths[0] / 2)))
        peak_indices = (left_peak_index, right_peak_index)
        peak_heights = (peak_heights[0] / 2, peak_heights[0] / 2)
        peak_widths = (peak_widths[0] / 2, peak_widths[0] / 2)
    elif len(peak_indices) == 0:
        # If no peaks have been found, just make a wild guess
        peak_indices = (len(x) // 3, 2 * len(x) // 3)
        peak_heights = (data_span, data_span)
        peak_widths = (x_spacing * 10, x_spacing * 10)

    estimate = {'height': np.asarray(peak_heights),
                'fwhm'  : np.asarray(peak_widths) * x_spacing,
                'center': np.asarray(x[np.asarray(peak_indices)])}
    limits = {'height': ((0, 2 * data_span),) * 2,
              'fwhm'  : ((x_spacing, x_span),) * 2,
              'center': ((min(x) - x_span / 2, max(x) + x_span / 2),) * 2}
    return estimate, limits


def estimate_triple_peaks(data, x, minimum_distance=None):
    # Find peaks along with width and amplitude estimation
    peak_indices, peak_heights, peak_widths = find_highest_peaks(data,
                                                                 peak_count=3,
                                                                 width=minimum_distance,
                                                                 height=0.05 * max(data))

    x_spacing = min(abs(np.ediff1d(x)))
    x_span = abs(x[-1] - x[0])
    data_span = abs(max(data) - min(data))

    # Replace missing peaks with sensible default value
    if len(peak_indices) == 2:
        # If just two peaks were found, assume it is three peaks overlapping and put one in the
        # middle
        middle_peak_index = abs(peak_indices[0] - peak_indices[1]) // 2
        middle_peak_width = np.mean(peak_widths)
        middle_peak_height = np.mean(peak_heights)
        peak_indices = (peak_indices[0], middle_peak_index, peak_indices[1])
        peak_heights = (peak_heights[0], middle_peak_height, peak_heights[1])
        peak_widths = (peak_widths[0], middle_peak_width, peak_widths[1])
    elif len(peak_indices) == 1:
        # If just one peak was found, assume it is three peaks overlapping and split it
        left_peak_index = max(0, int(round(peak_indices[0] - peak_widths[0] / 2)))
        right_peak_index = min(len(x) - 1, int(round(peak_indices[0] + peak_widths[0] / 2)))
        peak_indices = (left_peak_index, peak_indices[0], right_peak_index)
        peak_heights = (peak_heights[0] / 2, peak_heights[0] / 2, peak_heights[0] / 2)
        peak_widths = (peak_widths[0] / 2, peak_widths[0] / 2, peak_widths[0] / 2)
    elif len(peak_indices) == 0:
        # If no peaks have been found, just make a wild guess
        peak_indices = (len(x) // 4, len(x) // 2, 3 * len(x) // 4)
        peak_heights = (data_span, data_span, data_span)
        peak_widths = (x_spacing * 10, x_spacing * 10, x_spacing * 10)

    estimate = {'height': np.asarray(peak_heights),
                'fwhm'  : np.asarray(peak_widths) * x_spacing,
                'center': np.asarray(x[np.asarray(peak_indices)])}
    limits = {'height': ((0, 2 * data_span),) * 3,
              'fwhm'  : ((x_spacing, x_span),) * 3,
              'center': ((min(x) - x_span / 2, max(x) + x_span / 2),) * 3}
    return estimate, limits



# -*- coding: utf-8 -*-

"""
QWidget serving as main editor for the individual module config sections

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['ModuleEditorWidget']

from PySide2 import QtCore, QtWidgets
from typing import Optional, Mapping, Dict, Union, Any
from qudi.util.widgets.separator_lines import HorizontalLine
from qudi.tools.config_editor.module_finder import QudiModules
from qudi.tools.config_editor.module_widgets import LocalModuleConfigWidget
from qudi.tools.config_editor.module_widgets import RemoteModuleConfigWidget


class ModuleEditorWidget(QtWidgets.QStackedWidget):
    """
    """

    sigModuleRenamed = QtCore.Signal(str)

    def __init__(self,
                 qudi_modules: QudiModules,
                 parent: Optional[QtWidgets.QWidget] = None
                 ) -> None:
        super().__init__(parent=parent)
        self.setFrameShape(QtWidgets.QFrame.Shape.StyledPanel)

        assert isinstance(qudi_modules, QudiModules)
        self._qudi_modules = qudi_modules

        self.placeholder_label = QtWidgets.QLabel(
            'Please select a module to configure from the tree view.'
        )
        font = self.placeholder_label.font()
        font.setBold(True)
        font.setPointSize(font.pointSize() + 4)
        self.placeholder_label.setFont(font)
        self.placeholder_label.setAlignment(QtCore.Qt.AlignCenter)
        self.placeholder_label.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                             QtWidgets.QSizePolicy.Expanding)
        self.addWidget(self.placeholder_label)

        self._editor_layout = QtWidgets.QVBoxLayout()
        self._editor_layout.setStretch(2, 1)

        # Module name editor
        sub_layout = QtWidgets.QHBoxLayout()
        sub_layout.setStretch(1, 1)
        label = QtWidgets.QLabel('* Module Name:')
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.module_name_lineedit = QtWidgets.QLineEdit()
        self.module_name_lineedit.setPlaceholderText('Enter locally unique module name')
        self.module_name_lineedit.setFont(font)
        self.module_name_lineedit.textChanged.connect(self.sigModuleRenamed)
        sub_layout.addWidget(label)
        sub_layout.addWidget(self.module_name_lineedit)
        self._editor_layout.addLayout(sub_layout)

        # Add separator
        self._editor_layout.addWidget(HorizontalLine())

        editor_widget = QtWidgets.QWidget()
        editor_widget.setLayout(self._editor_layout)
        self.addWidget(editor_widget)

        self._current_editor = None

    @property
    def config(self) -> Union[None, Dict[str, Any]]:
        try:
            return self._current_editor.config
        except AttributeError:
            return None

    def set_config(self, config: Union[None, Dict[str, Any]]) -> None:
        try:
            self._current_editor.set_config(config)
        except AttributeError:
            pass

    @property
    def module_name(self) -> str:
        return self.module_name_lineedit.text()

    def set_module_name(self, name: str) -> None:
        self.module_name_lineedit.blockSignals(True)
        try:
            self.module_name_lineedit.setText(name)
        finally:
            self.module_name_lineedit.blockSignals(False)

    def open_remote_module(self,
                           name: Optional[str] = None,
                           config: Optional[Mapping[str, Union[str, None]]] = None
                           ) -> None:
        if self._current_editor is not None:
            self.close_editor()

        self._current_editor = RemoteModuleConfigWidget(config=config)
        self._editor_layout.addWidget(self._current_editor)
        self.set_module_name(name if name else '')
        self.setCurrentIndex(1)

    def open_local_module(self,
                          module_class: str,
                          named_modules: Mapping[str, str],
                          name: Optional[str] = None,
                          config: Optional[Dict[str, Union[str, bool, Dict[str, str], Dict[str, Any]]]] = None,
                          ) -> None:
        if self._current_editor is not None:
            self.close_editor()

        connectors = self._qudi_modules.module_connectors(module_class)
        config_options = self._qudi_modules.module_config_options(module_class)
        valid_connector_targets = self._qudi_modules.module_connector_targets(module_class)
        self._current_editor = LocalModuleConfigWidget(
            module_class=module_class,
            config_options=config_options,
            connectors=connectors,
            valid_connector_targets=valid_connector_targets,
            named_modules=named_modules,
            config=config
        )
        self._editor_layout.addWidget(self._current_editor)
        self.set_module_name(name if name else '')
        self.setCurrentIndex(1)

    def close_editor(self):
        if self._current_editor is None:
            return
        self.setCurrentIndex(0)
        self._editor_layout.removeWidget(self._current_editor)
        self._current_editor.setParent(None)
        self._current_editor.deleteLater()
        self._current_editor = None



# -*- coding: utf-8 -*-
"""
This file contains a custom module widget for the Qudi manager GUI.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

import os
from PySide2 import QtCore, QtGui, QtWidgets
from qudi.util.paths import get_artwork_dir
from qudi.util.mutex import Mutex


class ModuleFrameWidget(QtWidgets.QWidget):
    """
    Custom module QWidget for the Qudi main GUI.
    """
    sigActivateClicked = QtCore.Signal(str)
    sigDeactivateClicked = QtCore.Signal(str)
    sigReloadClicked = QtCore.Signal(str)
    sigCleanupClicked = QtCore.Signal(str)

    def __init__(self, *args, module_name=None, **kwargs):
        super().__init__(*args, **kwargs)

        # Create QToolButtons
        self.cleanup_button = QtWidgets.QToolButton()
        self.cleanup_button.setObjectName('cleanupButton')
        self.deactivate_button = QtWidgets.QToolButton()
        self.deactivate_button.setObjectName('deactivateButton')
        self.reload_button = QtWidgets.QToolButton()
        self.reload_button.setObjectName('reloadButton')

        # Set icons for QToolButtons
        icon_path = os.path.join(get_artwork_dir(), 'icons')
        self.cleanup_button.setIcon(QtGui.QIcon(os.path.join(icon_path, 'edit-clear')))
        self.deactivate_button.setIcon(QtGui.QIcon(os.path.join(icon_path, 'edit-delete')))
        self.reload_button.setIcon(QtGui.QIcon(os.path.join(icon_path, 'view-refresh')))

        # Create activation pushbutton
        self.activate_button = QtWidgets.QPushButton('load/activate <module_name>')
        self.activate_button.setObjectName('loadButton')
        self.activate_button.setCheckable(True)
        self.activate_button.setMinimumWidth(200)
        self.activate_button.setSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding,
                                           QtWidgets.QSizePolicy.Fixed)

        # Create status label
        self.status_label = QtWidgets.QLabel('Module status goes here...')
        self.status_label.setObjectName('statusLabel')

        # Set tooltips
        self.cleanup_button.setToolTip('Clean up module status file')
        self.deactivate_button.setToolTip('Deactivate module')
        self.reload_button.setToolTip('Reload module')
        self.activate_button.setToolTip('Load this module and all its dependencies')
        self.status_label.setToolTip('Displays module status information')

        # Combine all widgets in a layout and set as main layout
        layout = QtWidgets.QGridLayout()
        layout.addWidget(self.activate_button, 0, 0)
        layout.addWidget(self.reload_button, 0, 1)
        layout.addWidget(self.deactivate_button, 0, 2)
        layout.addWidget(self.cleanup_button, 0, 3)
        layout.addWidget(self.status_label, 1, 0, 1, 4)
        self.setLayout(layout)

        self._module_name = ''
        if module_name:
            self.set_module_name(module_name)

        self.activate_button.clicked.connect(self.activate_clicked)
        self.deactivate_button.clicked.connect(self.deactivate_clicked)
        self.reload_button.clicked.connect(self.reload_clicked)
        self.cleanup_button.clicked.connect(self.cleanup_clicked)
        return

    def set_module_name(self, name):
        if name:
            self.activate_button.setText('Load {0}'.format(name))
            self._module_name = name

    def set_module_state(self, state):
        if state == 'not loaded':
            self.activate_button.setText('Load {0}'.format(self._module_name))
            self.cleanup_button.setEnabled(True)
            self.deactivate_button.setEnabled(False)
            self.reload_button.setEnabled(False)
            if self.activate_button.isChecked():
                self.activate_button.setChecked(False)
        elif state == 'deactivated':
            self.activate_button.setText('Activate {0}'.format(self._module_name))
            self.cleanup_button.setEnabled(True)
            self.deactivate_button.setEnabled(False)
            self.reload_button.setEnabled(True)
            if self.activate_button.isChecked():
                self.activate_button.setChecked(False)
        else:
            self.activate_button.setText(self._module_name)
            self.cleanup_button.setEnabled(False)
            self.deactivate_button.setEnabled(True)
            self.reload_button.setEnabled(True)
            if not self.activate_button.isChecked():
                self.activate_button.setChecked(True)
        self.status_label.setText('Module is {0}'.format(state))

    def set_module_app_data(self, exists):
        self.cleanup_button.setEnabled(exists)

    @QtCore.Slot()
    def activate_clicked(self):
        self.sigActivateClicked.emit(self._module_name)

    @QtCore.Slot()
    def deactivate_clicked(self):
        self.sigDeactivateClicked.emit(self._module_name)

    @QtCore.Slot()
    def cleanup_clicked(self):
        self.sigCleanupClicked.emit(self._module_name)

    @QtCore.Slot()
    def reload_clicked(self):
        self.sigReloadClicked.emit(self._module_name)


class ModuleListModel(QtCore.QAbstractListModel):
    """
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._lock = Mutex()
        self._module_states = dict()
        self._module_app_data = dict()
        self._module_names = list()

    def rowCount(self, parent):
        return len(self._module_names)

    def data(self, index, role):
        if not index.isValid():
            return
        row = index.row()
        if row >= len(self._module_names):
            return
        name = self._module_names[row]
        state = self._module_states[name]
        app_data = self._module_app_data[name]
        if role == QtCore.Qt.DisplayRole:
            return name, state, app_data

    def flags(self, index):
        return QtCore.Qt.ItemNeverHasChildren | QtCore.Qt.ItemIsEnabled

    def append_module(self, name, state, app_data):
        with self._lock:
            if name in self._module_states:
                raise RuntimeError(f'Module with name "{name}" already present in ModuleListModel.')
            self.beginInsertRows(len(self._module_names))
            self._module_names.append(name)
            self._module_states[name] = state
            self._module_app_data[name] = app_data
            self.endInsertRows()

    def remove_module(self, name):
        with self._lock:
            if name not in self._module_states:
                return
            row = self._module_names.index(name)
            self.beginRemoveRows(row, row + 1)
            del self._module_names[row]
            del self._module_states[name]
            del self._module_app_data[name]
            self.endRemoveRows()

    def reset_modules(self, state_dict, app_data_dict):
        if set(state_dict) != set(app_data_dict):
            raise RuntimeError('state_dict and app_data_dict must contain exactly the same keys.')
        with self._lock:
            self.beginResetModel()
            self._module_states = state_dict.copy()
            self._module_app_data = app_data_dict.copy()
            self._module_names = list(state_dict)
            self.endResetModel()

    def change_module_state(self, name, state):
        with self._lock:
            if name not in self._module_states:
                raise RuntimeError(
                    f'Can not change module state in ModuleListModel. No module by the name '
                    f'"{name}" found.'
                )
            self._module_states[name] = state
            row = self._module_names.index(name)
            self.dataChanged.emit(self.createIndex(row, 0),
                                  self.createIndex(row + 1, 0),
                                  (QtCore.Qt.DisplayRole,))

    def change_app_data(self, name, exists):
        with self._lock:
            if name not in self._module_app_data:
                raise RuntimeError(
                    f'Can not change module app status in ModuleListModel. No module by the name '
                    f'"{name}" found.'
                )
            self._module_app_data[name] = exists
            row = self._module_names.index(name)
            self.dataChanged.emit(self.createIndex(row, 0), self.createIndex(row, 0))


class ModuleListItemDelegate(QtWidgets.QStyledItemDelegate):
    """
    """
    sigActivateClicked = QtCore.Signal(str)
    sigDeactivateClicked = QtCore.Signal(str)
    sigReloadClicked = QtCore.Signal(str)
    sigCleanupClicked = QtCore.Signal(str)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.render_widget = ModuleFrameWidget()
        self.__origin = QtCore.QPoint()

    def createEditor(self, parent, option, index):
        widget = ModuleFrameWidget(parent=parent)
        # Found no other way to pefectly match editor and rendered item view (using paint())
        widget.setContentsMargins(2, 2, 2, 2)
        widget.sigActivateClicked.connect(self.sigActivateClicked)
        widget.sigDeactivateClicked.connect(self.sigDeactivateClicked)
        widget.sigReloadClicked.connect(self.sigReloadClicked)
        widget.sigCleanupClicked.connect(self.sigCleanupClicked)
        return widget

    def setEditorData(self, editor, index):
        data = index.data()
        if data:
            editor.set_module_name(data[0])
            editor.set_module_state(data[1])
            editor.set_module_app_data(data[2])

    def setModelData(self, editor, model, index):
        pass

    def sizeHint(self, option=None, index=None):
        return self.render_widget.sizeHint()

    def paint(self, painter, option, index):
        """
        """
        name, state, app_data = index.data()
        self.render_widget.set_module_name(name)
        self.render_widget.set_module_state(state)
        self.render_widget.set_module_app_data(app_data)
        self.render_widget.setGeometry(option.rect)
        painter.save()
        painter.translate(option.rect.topLeft())
        self.render_widget.render(painter, self.__origin)
        painter.restore()


class ModuleListView(QtWidgets.QListView):
    """
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setMouseTracking(True)
        delegate = ModuleListItemDelegate()
        self.setItemDelegate(delegate)
        self.setMinimumWidth(delegate.sizeHint().width())
        self.setUniformItemSizes(True)
        self.setContentsMargins(0, 0, 0, 0)
        self.setSpacing(1)
        self.previous_index = QtCore.QModelIndex()

    def mouseMoveEvent(self, event):
        index = self.indexAt(event.pos())
        if index != self.previous_index:
            if self.previous_index.isValid():
                self.closePersistentEditor(self.previous_index)
            if index.isValid():
                self.openPersistentEditor(index)
            self.previous_index = index

    def leaveEvent(self, event):
        if self.previous_index.isValid():
            self.closePersistentEditor(self.previous_index)
        self.previous_index = QtCore.QModelIndex()


class ModuleWidget(QtWidgets.QTabWidget):
    """
    """
    sigActivateModule = QtCore.Signal(str)
    sigDeactivateModule = QtCore.Signal(str)
    sigCleanupModule = QtCore.Signal(str)
    sigReloadModule = QtCore.Signal(str)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        self.list_models = {'gui'     : ModuleListModel(),
                            'logic'   : ModuleListModel(),
                            'hardware': ModuleListModel()}
        self.list_views = {'gui'     : ModuleListView(),
                           'logic'   : ModuleListView(),
                           'hardware': ModuleListView()}
        self.addTab(self.list_views['gui'], 'GUI')
        self.addTab(self.list_views['logic'], 'Logic')
        self.addTab(self.list_views['hardware'], 'Hardware')
        for base, view in self.list_views.items():
            view.setModel(self.list_models[base])
            delegate = view.itemDelegate()
            delegate.sigActivateClicked.connect(self.sigActivateModule)
            delegate.sigDeactivateClicked.connect(self.sigDeactivateModule)
            delegate.sigReloadClicked.connect(self.sigReloadModule)
            delegate.sigCleanupClicked.connect(self.sigCleanupModule)

    @QtCore.Slot(dict)
    def update_modules(self, modules_dict):
        for base, model in self.list_models.items():
            model.reset_modules(
                {name: mod.state for name, mod in modules_dict.items() if mod.module_base == base},
                {name: mod.has_app_data for name, mod in modules_dict.items() if
                 mod.module_base == base}
            )
        return

    @QtCore.Slot(str, str, str)
    def update_module_state(self, base, name, state):
        self.list_models[base].change_module_state(name, state)

    @QtCore.Slot(str, str, bool)
    def update_module_app_data(self, base, name, exists):
        self.list_models[base].change_app_data(name, exists)



# -*- coding: utf-8 -*-

"""
Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['RectangleROI']

from math import isinf
from typing import Union, Tuple, Optional, Sequence, List
from PySide2 import QtCore
from pyqtgraph import ROI


class RectangleROI(ROI):
    """
    """
    def __init__(self,
                 pos=(0, 0),
                 size=(1, 1),
                 bounds=None,
                 apply_bounds_to_center=False,
                 parent=None,
                 pen=None,
                 hoverPen=None,
                 handlePen=None,
                 handleHoverPen=None,
                 movable=True,
                 resizable=True,
                 aspectLocked=False
                 ) -> None:
        ROI.__init__(self,
                     (0, 0),
                     size=(1, 1),
                     angle=0,
                     invertible=True,
                     maxBounds=None,
                     scaleSnap=False,
                     translateSnap=False,
                     rotateSnap=False,
                     parent=parent,
                     pen=pen,
                     hoverPen=hoverPen,
                     handlePen=handlePen,
                     handleHoverPen=handleHoverPen,
                     movable=movable,
                     rotatable=False,
                     resizable=resizable,
                     removable=False,
                     aspectLocked=aspectLocked)
        self.__center_position = (0, 0)
        self.__norm_size = (1, 1)
        self.__min_norm_size = (0, 0)
        self.__start_pos = (0, 0)
        self._apply_bounds_to_center = bool(apply_bounds_to_center)
        self._bounds = self.normalize_bounds(bounds)
        self.set_area(position=pos, size=size)

    @property
    def min_size(self) -> Tuple[float, float]:
        return self.__min_norm_size

    def set_min_size(self, size: Union[None, Tuple[float, float]]) -> None:
        self.__min_norm_size = (0, 0) if size is None else (abs(size[0]), abs(size[1]))
        self.set_area(size=self.__norm_size)

    @property
    def area(self) -> Tuple[Tuple[float, float], Tuple[float, float]]:
        return self.__center_position, self.__norm_size

    def set_area(self,
                 position: Optional[Tuple[float, float]] = None,
                 size: Optional[Tuple[float, float]] = None
                 ) -> None:
        if position is not None:
            constr_size = (max(self.__norm_size[0], self.__min_norm_size[0]),
                           max(self.__norm_size[1], self.__min_norm_size[1]))
            self.setPos(QtCore.QPointF(position[0] - constr_size[0] / 2,
                                       position[1] + constr_size[1] / 2),
                        update=False,
                        finish=False)
            self.__center_position = (position[0], position[1])
        if size is not None:
            size = (abs(size[0]), abs(size[1]))
            constr_size = (max(size[0], self.__min_norm_size[0]),
                           max(size[1], self.__min_norm_size[1]))
            self.setSize(QtCore.QPointF(constr_size[0], -constr_size[1]),
                         center=(0.5, 0.5),
                         update=False,
                         finish=False)
            self.__norm_size = size
        if position is not None and size is not None:
            self._clip_area(update=True, finish=True)

    @property
    def bounds(self) -> List[Tuple[Union[None, float], Union[None, float]]]:
        return self._bounds.copy()

    def set_bounds(self,
                   bounds: Union[None, Sequence[Tuple[Union[None, float], Union[None, float]]]]
                   ) -> None:
        self._bounds = self.normalize_bounds(bounds)
        self._clip_area(update=True, finish=True)

    def _clip_area(self, update: Optional[bool] = True, finish: Optional[bool] = True) -> None:
        position = list(self.__center_position)
        size = list(self.__norm_size)
        x_min, x_max = self._bounds[0]
        y_min, y_max = self._bounds[1]
        if self._apply_bounds_to_center:
            if (x_min is not None) and (position[0] < x_min):
                position[0] = x_min
            elif (x_max is not None) and (position[0] > x_max):
                position[0] = x_max
            if (y_min is not None) and (position[1] < y_min):
                position[1] = y_min
            elif (y_max is not None) and (position[1] > y_max):
                position[1] = y_max
        else:
            left = position[0] - size[0] / 2
            right = position[0] + size[0] / 2
            top = position[1] + size[1] / 2
            bottom = position[1] - size[1] / 2
            if (x_min is not None) and (left < x_min):
                position[0] = x_min + size[0] / 2
            elif (x_max is not None) and (right > x_max):
                position[0] = x_max - size[0] / 2
            if (y_min is not None) and (bottom < y_min):
                position[1] = y_min + size[1] / 2
            elif (y_max is not None) and (top > y_max):
                position[1] = y_max - size[1] / 2
        translate = (position[0] - self.__center_position[0],
                     position[1] - self.__center_position[1])
        self.__center_position = tuple(position)
        self.translate(translate, update=update, finish=finish)

    @staticmethod
    def normalize_bounds(bounds: Union[None, Sequence[Tuple[Union[None, float], Union[None, float]]]]
                         ) -> List[Tuple[Union[None, float], Union[None, float]]]:
        if bounds is None:
            bounds = [(None, None), (None, None)]
        else:
            bounds = [list(span) for span in bounds]
            # Replace inf values by None
            for span in bounds:
                for ii, val in enumerate(span):
                    try:
                        if isinf(val):
                            span[ii] = None
                    except TypeError:
                        pass
            # Sort spans in ascending order
            try:
                bounds[0] = tuple(sorted(bounds[0]))
            except TypeError:
                bounds[0] = tuple(bounds[0])
            try:
                bounds[1] = tuple(sorted(bounds[1]))
            except TypeError:
                bounds[1] = tuple(bounds[1])
        return bounds

    def checkPointMove(self, handle, pos, modifiers):
        if not self._apply_bounds_to_center:
            pos = self.mapSceneToParent(pos)
            x_min, x_max = self._bounds[0]
            y_min, y_max = self._bounds[1]
            if (x_min is not None) and pos.x() < x_min:
                return False
            if (x_max is not None) and pos.x() > x_max:
                return False
            if (y_min is not None) and pos.y() < y_min:
                return False
            if (y_max is not None) and pos.y() > y_max:
                return False
        return True

    def cancelMove(self) -> None:
        if self.isMoving:
            self.__center_position = self.__start_pos
        return super().cancelMove()

    def mouseDragEvent(self, ev) -> None:
        if not ev.isAccepted():
            if self.translatable and ev.button() == QtCore.Qt.LeftButton and ev.modifiers() == QtCore.Qt.NoModifier:
                is_start = ev.isStart()
                is_finish = ev.isFinish()
                ev.accept()
                if is_start:
                    self.setSelected(True)
                    self._moveStarted()
                    self.__start_pos = self.__center_position
                if self.isMoving:
                    total_move = self.mapToParent(ev.pos()) - self.mapToParent(ev.buttonDownPos())
                    self.__center_position = (self.__start_pos[0] + total_move.x(),
                                              self.__start_pos[1] + total_move.y())
                    new_pos = self.preMoveState['pos'] + total_move
                    self.setPos(new_pos, update=False, finish=False)
                    self._clip_area(update=True, finish=False)
                if is_finish:
                    self._moveFinished()
            else:
                ev.ignore()



# -*- coding: utf-8 -*-
"""
This file contains the Qudi console app class.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

import os
import weakref
import platform
from PySide2 import QtCore, QtGui, QtWidgets
from qudi.core.gui.main_gui.main_gui import QudiMainGui
from qudi.core.modulemanager import ModuleManager
from qudi.util.paths import get_artwork_dir
from qudi.core.logger import get_logger

try:
    import pyqtgraph as pg
except ImportError:
    pg = None

logger = get_logger(__name__)


class SystemTrayIcon(QtWidgets.QSystemTrayIcon):
    """Tray icon class subclassing QSystemTrayIcon for custom functionality.
    """

    def __init__(self):
        """Tray icon constructor.
        Adds all the appropriate menus and actions.
        """
        super().__init__()
        self._actions = dict()
        self.setIcon(QtWidgets.QApplication.instance().windowIcon())
        self.right_menu = QtWidgets.QMenu('Quit')
        self.left_menu = QtWidgets.QMenu('Manager')

        iconpath = os.path.join(get_artwork_dir(), 'icons')
        self.managericon = QtGui.QIcon()
        self.managericon.addFile(os.path.join(iconpath, 'go-home'), QtCore.QSize(16, 16))
        self.managerAction = QtWidgets.QAction(self.managericon, 'Manager', self.left_menu)

        self.exiticon = QtGui.QIcon()
        self.exiticon.addFile(os.path.join(iconpath, 'application-exit'), QtCore.QSize(16, 16))
        self.quitAction = QtWidgets.QAction(self.exiticon, 'Quit', self.right_menu)

        self.restarticon = QtGui.QIcon()
        self.restarticon.addFile(os.path.join(iconpath, 'view-refresh'), QtCore.QSize(16, 16))
        self.restartAction = QtWidgets.QAction(self.restarticon, 'Restart', self.right_menu)

        self.left_menu.addAction(self.managerAction)
        self.left_menu.addSeparator()

        self.right_menu.addAction(self.quitAction)
        self.right_menu.addAction(self.restartAction)
        self.setContextMenu(self.right_menu)

        self.activated.connect(self.handle_activation)

    @QtCore.Slot(QtWidgets.QSystemTrayIcon.ActivationReason)
    def handle_activation(self, reason):
        """Click handler.
        This method is called when the tray icon is left-clicked.
        It opens a menu at the position of the left click.

        @param reason: reason that caused the activation
        """
        if reason == self.Trigger:
            self.left_menu.exec_(QtGui.QCursor.pos())

    def add_action(self, label, callback, icon=None):
        if label in self._actions:
            raise ValueError(f'Action "{label}" already exists in system tray.')

        if not isinstance(icon, QtGui.QIcon):
            icon = QtGui.QIcon()
            iconpath = os.path.join(get_artwork_dir(), 'icons')
            icon.addFile(os.path.join(iconpath, 'go-next'))

        action = QtWidgets.QAction(label)
        action.setIcon(icon)
        action.triggered.connect(callback)
        self.left_menu.addAction(action)
        self._actions[label] = action

    def remove_action(self, label):
        action = self._actions.pop(label, None)
        if action is not None:
            action.triggered.disconnect()
            self.left_menu.removeAction(action)


class Gui(QtCore.QObject):
    """Set up all necessary GUI elements, like application icons, themes, etc.
    """

    _instance = None

    _sigPopUpMessage = QtCore.Signal(str, str)
    _sigBalloonMessage = QtCore.Signal(str, str, object, object)

    def __new__(cls, *args, **kwargs):
        if cls._instance is None or cls._instance() is None:
            obj = super().__new__(cls, *args, **kwargs)
            cls._instance = weakref.ref(obj)
            return obj
        raise RuntimeError(
            'Gui is a singleton. Please use Gui.instance() to get a reference to the already '
            'created instance.'
        )

    def __init__(self, qudi_instance, stylesheet_path=None, theme=None, use_opengl=False):
        if theme is None:
            theme = 'qudiTheme'

        super().__init__()

        app = QtWidgets.QApplication.instance()
        if app is None:
            raise RuntimeError('No Qt GUI app running (no QApplication instance).')

        app.setQuitOnLastWindowClosed(False)

        self._init_app_icon()
        self.set_theme(theme)
        if stylesheet_path is not None:
            self.set_style_sheet(stylesheet_path)
        self.system_tray_icon = SystemTrayIcon()

        self._sigPopUpMessage.connect(self.pop_up_message, QtCore.Qt.QueuedConnection)
        self._sigBalloonMessage.connect(self.balloon_message, QtCore.Qt.QueuedConnection)

        self._configure_pyqtgraph(use_opengl)
        self.main_gui_module = QudiMainGui(qudi_main_weakref=weakref.ref(qudi_instance),
                                           name='qudi_main_gui')
        self.system_tray_icon.managerAction.triggered.connect(self.activate_main_gui,
                                                              QtCore.Qt.QueuedConnection)
        self.system_tray_icon.quitAction.triggered.connect(qudi_instance.quit,
                                                           QtCore.Qt.QueuedConnection)
        self.system_tray_icon.restartAction.triggered.connect(qudi_instance.restart,
                                                              QtCore.Qt.QueuedConnection)
        qudi_instance.module_manager.sigModuleStateChanged.connect(self._tray_module_action_changed)
        self.show_system_tray_icon()

    @classmethod
    def instance(cls):
        if cls._instance is None:
            return None
        return cls._instance()

    @staticmethod
    def _init_app_icon():
        """Set up the Qudi application icon.
        """
        app_icon = QtGui.QIcon(os.path.join(get_artwork_dir(), 'logo', 'logo-qudi.svg'))
        QtWidgets.QApplication.instance().setWindowIcon(app_icon)

    @staticmethod
    def _configure_pyqtgraph(use_opengl=False):
        # Configure pyqtgraph (if present)
        if pg is not None:
            # test setting background of pyqtgraph
            testwidget = QtWidgets.QWidget()
            testwidget.ensurePolished()
            bgcolor = testwidget.palette().color(QtGui.QPalette.Normal, testwidget.backgroundRole())
            # set manually the background color in hex code according to our color scheme:
            pg.setConfigOption('background', bgcolor)
            # experimental opengl usage
            pg.setConfigOption('useOpenGL', use_opengl)

    @staticmethod
    def set_theme(theme):
        """
        Set icon theme for qudi app.

        @param str theme: qudi theme name
        """
        # Make icons work on non-X11 platforms, set custom theme
        # if not sys.platform.startswith('linux') and not sys.platform.startswith('freebsd'):
        #
        # To enable the use of custom action icons, for now the above if statement has been
        # removed and the QT theme is being set to our artwork/icons folder for
        # all OSs.
        themepaths = QtGui.QIcon.themeSearchPaths()
        themepaths.append(os.path.join(get_artwork_dir(), 'icons'))
        QtGui.QIcon.setThemeSearchPaths(themepaths)
        QtGui.QIcon.setThemeName(theme)

    @staticmethod
    def set_style_sheet(stylesheet_path):
        """
        Set qss style sheet for application.

        @param str stylesheet_path: path to style sheet file
        """
        try:
            if not os.path.exists(stylesheet_path):
                stylesheet_path = os.path.join(get_artwork_dir(), 'styles', stylesheet_path)

            with open(stylesheet_path, 'r') as stylesheetfile:
                stylesheet = stylesheetfile.read()

            if stylesheet_path.endswith('qdark.qss'):
                path = os.path.join(os.path.dirname(stylesheet_path), 'qdark').replace('\\', '/')
                stylesheet = stylesheet.replace('{qdark}', path)

            # see issue #12 on qdarkstyle github
            if platform.system().lower() == 'darwin' and stylesheet_path.endswith('qdark.qss'):
                mac_fix = '''
                QDockWidget::title
                {
                    background-color: #31363b;
                    text-align: center;
                    height: 12px;
                }
                '''
                stylesheet += mac_fix
            QtWidgets.QApplication.instance().setStyleSheet(stylesheet)
        except:
            logger.exception('Exception while setting qudi stylesheet:')

    @staticmethod
    def close_windows():
        """Close all application windows.
        """
        QtWidgets.QApplication.instance().closeAllWindows()

    def activate_main_gui(self):
        if QtCore.QThread.currentThread() is not self.thread():
            QtCore.QMetaObject.invokeMethod(self,
                                            'activate_main_gui',
                                            QtCore.Qt.BlockingQueuedConnection)
            return

        if self.main_gui_module.module_state() != 'deactivated':
            self.main_gui_module.show()
            return

        logger.info('Activating main GUI module...')
        print('> Activating main GUI module...')

        self.main_gui_module.module_state.activate()
        QtWidgets.QApplication.instance().processEvents()

    def deactivate_main_gui(self):
        if QtCore.QThread.currentThread() is not self.thread():
            QtCore.QMetaObject.invokeMethod(self,
                                            'deactivate_main_gui',
                                            QtCore.Qt.BlockingQueuedConnection)
            return

        if self.main_gui_module.module_state() == 'deactivated':
            return

        self.main_gui_module.module_state.deactivate()
        QtWidgets.QApplication.instance().processEvents()

    def show_system_tray_icon(self):
        """Show system tray icon.
        """
        self.system_tray_icon.show()

    def hide_system_tray_icon(self):
        """Hide system tray icon.
        """
        self.system_tray_icon.hide()

    def close_system_tray_icon(self):
        """
        Kill and delete system tray icon. Tray icon will be lost until Gui.__init__ is called again.
        """
        self.hide_system_tray_icon()
        self.system_tray_icon.quitAction.triggered.disconnect()
        self.system_tray_icon.restartAction.triggered.disconnect()
        self.system_tray_icon.managerAction.triggered.disconnect()
        self.system_tray_icon = None

    def system_tray_notification_bubble(self, title, message, time=None, icon=None):
        """
        Helper method to invoke balloon messages in the system tray by calling
        QSystemTrayIcon.showMessage.

        @param str title: The notification title of the balloon
        @param str message: The message to be shown in the balloon
        @param float time: optional, The lingering time of the balloon in seconds
        @param QIcon icon: optional, an icon to be used in the balloon. "None" will use OS default.
        """
        if icon is None:
            icon = QtGui.QIcon()
        if time is None:
            time = 15
        self.system_tray_icon.showMessage(title, message, icon, int(round(time * 1000)))

    def prompt_shutdown(self, modules_locked=True):
        """Display a dialog, asking the user to confirm shutdown.
        """
        if modules_locked:
            msg = 'Some qudi modules are locked right now.\n' \
                  'Do you really want to quit and force modules to deactivate?'
        else:
            msg = 'Do you really want to quit?'

        result = QtWidgets.QMessageBox.question(self.main_gui_module.mw,
                                                'Qudi: Quit?',
                                                msg,
                                                QtWidgets.QMessageBox.Yes,
                                                QtWidgets.QMessageBox.No)
        return result == QtWidgets.QMessageBox.Yes

    def prompt_restart(self, modules_locked=True):
        """Display a dialog, asking the user to confirm restart.
        """
        if modules_locked:
            msg = 'Some qudi modules are locked right now.\n' \
                  'Do you really want to restart and force modules to deactivate?'
        else:
            msg = 'Do you really want to restart?'

        result = QtWidgets.QMessageBox.question(self.main_gui_module.mw,
                                                'Qudi: Restart?',
                                                msg,
                                                QtWidgets.QMessageBox.Yes,
                                                QtWidgets.QMessageBox.No)
        return result == QtWidgets.QMessageBox.Yes

    @QtCore.Slot(str, str)
    def pop_up_message(self, title, message):
        """
        Slot prompting a dialog window with a message and an OK button to dismiss it.

        @param str title: The window title of the dialog
        @param str message: The message to be shown in the dialog window
        """
        if not isinstance(title, str):
            logger.error('pop-up message title must be str type')
            return
        if not isinstance(message, str):
            logger.error('pop-up message must be str type')
            return
        if self.thread() is not QtCore.QThread.currentThread():
            self._sigPopUpMessage.emit(title, message)
            return
        QtWidgets.QMessageBox.information(None, title, message, QtWidgets.QMessageBox.Ok)
        return

    @QtCore.Slot(str, str, object, object)
    def balloon_message(self, title, message, time=None, icon=None):
        """
        Slot prompting a balloon notification from the system tray icon.

        @param str title: The notification title of the balloon
        @param str message: The message to be shown in the balloon
        @param float time: optional, The lingering time of the balloon in seconds
        @param QIcon icon: optional, an icon to be used in the balloon. "None" will use OS default.
        """
        if not self.system_tray_icon.supportsMessages():
            logger.warning('{0}:\n{1}'.format(title, message))
            return
        if self.thread() is not QtCore.QThread.currentThread():
            self._sigBalloonMessage.emit(title, message, time, icon)
            return
        self.system_tray_notification_bubble(title, message, time=time, icon=icon)
        return

    @QtCore.Slot(str, str, str)
    def _tray_module_action_changed(self, base, module_name, state):
        if self.system_tray_icon and base == 'gui':
            if state == 'deactivated':
                self.system_tray_icon.remove_action(module_name)
            else:
                mod_manager = ModuleManager.instance()
                try:
                    module_inst = mod_manager[module_name].instance
                except KeyError:
                    return
                self.system_tray_icon.add_action(module_name, module_inst.show)



# -*- coding: utf-8 -*-

"""
ToDo: Document

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ('is_fit_model', 'get_all_fit_models', 'FitConfiguration', 'FitConfigurationsModel',
           'FitContainer')

import importlib
import logging
import inspect
import lmfit
import numpy as np
from PySide2 import QtCore
from typing import Iterable, Optional, Mapping, Union

import qudi.util.fit_models as _fit_models_ns
from qudi.util.mutex import Mutex
from qudi.util.units import create_formatted_output
from qudi.util.helpers import iter_modules_recursive
from qudi.util.fit_models.model import FitModelBase


_log = logging.getLogger(__name__)


def is_fit_model(cls):
    return inspect.isclass(cls) and issubclass(cls, FitModelBase) and (cls is not FitModelBase)


# Upon import of this module the global attribute _fit_models is initialized with a dict
# containing all importable fit model objects with names as keys.
_fit_models = dict()
for mod_finder in iter_modules_recursive(_fit_models_ns.__path__, _fit_models_ns.__name__ + '.'):
    try:
        _fit_models.update(
            {name: cls for name, cls in
             inspect.getmembers(importlib.import_module(mod_finder.name), is_fit_model)}
        )
    except:
        _log.exception(
            f'Exception while importing qudi.util.fit_models sub-module "{mod_finder.name}":'
        )


def get_all_fit_models():
    return _fit_models.copy()


class FitConfiguration:
    """
    """

    def __init__(self, name, model, estimator=None, custom_parameters=None):
        assert isinstance(name, str), 'FitConfiguration name must be str type.'
        assert name, 'FitConfiguration name must be non-empty string.'
        assert model in _fit_models, f'Invalid fit model name encountered: "{model}".'
        assert name != 'No Fit', '"No Fit" is a reserved name for fit configs. Choose another.'

        self._name = name
        self._model = model
        self._estimator = None
        self._custom_parameters = None
        self.estimator = estimator
        self.custom_parameters = custom_parameters

    @property
    def name(self):
        return self._name

    @property
    def model(self):
        return self._model

    @property
    def estimator(self):
        return self._estimator

    @estimator.setter
    def estimator(self, value):
        if value is not None:
            assert value in self.available_estimators, \
                f'Invalid fit model estimator encountered: "{value}"'
        self._estimator = value

    @property
    def available_estimators(self):
        return tuple(_fit_models[self._model]().estimators)

    @property
    def default_parameters(self):
        params = _fit_models[self._model]().make_params()
        return lmfit.Parameters() if params is None else params

    @property
    def custom_parameters(self):
        return self._custom_parameters.copy() if self._custom_parameters is not None else None

    @custom_parameters.setter
    def custom_parameters(self, value):
        if value is not None:
            default_params = self.default_parameters
            invalid = set(value).difference(default_params)
            assert not invalid, f'Invalid model parameters encountered: {invalid}'
            assert isinstance(value, lmfit.Parameters), \
                'Property custom_parameters must be of type <lmfit.Parameters>.'
        self._custom_parameters = value.copy() if value is not None else None

    def to_dict(self):
        return {
            'name': self._name,
            'model': self._model,
            'estimator': self._estimator,
            'custom_parameters': None if self._custom_parameters is None else self._custom_parameters.dumps()
        }

    @classmethod
    def from_dict(cls, dict_repr):
        assert set(dict_repr) == {'name', 'model', 'estimator', 'custom_parameters'}
        if isinstance(dict_repr['custom_parameters'], str):
            dict_repr['custom_parameters'] = lmfit.Parameters().loads(
                dict_repr['custom_parameters']
            )
        return cls(**dict_repr)


class FitConfigurationsModel(QtCore.QAbstractListModel):
    """
    """

    sigFitConfigurationsChanged = QtCore.Signal(tuple)

    def __init__(self, *args, configurations=None, **kwargs):
        assert (configurations is None) or all(isinstance(c, FitConfiguration) for c in configurations)
        super().__init__(*args, **kwargs)
        self._fit_configurations = list() if configurations is None else list(configurations)

    @property
    def model_names(self):
        return tuple(_fit_models)

    @property
    def model_estimators(self):
        return {name: tuple(model().estimators) for name, model in _fit_models.items()}

    @property
    def model_default_parameters(self):
        return {name: model().make_params() for name, model in _fit_models.items()}

    @property
    def configuration_names(self):
        return tuple(fc.name for fc in self._fit_configurations)

    @property
    def configurations(self):
        return self._fit_configurations.copy()

    @QtCore.Slot(str, str)
    def add_configuration(self, name, model):
        assert name not in self.configuration_names, f'Fit config "{name}" already defined.'
        assert name != 'No Fit', '"No Fit" is a reserved name for fit configs. Choose another.'
        config = FitConfiguration(name, model)
        new_row = len(self._fit_configurations)
        self.beginInsertRows(self.createIndex(new_row, 0), new_row, new_row)
        self._fit_configurations.append(config)
        self.endInsertRows()
        self.sigFitConfigurationsChanged.emit(self.configuration_names)

    @QtCore.Slot(str)
    def remove_configuration(self, name):
        try:
            row_index = self.configuration_names.index(name)
        except ValueError:
            return
        self.beginRemoveRows(self.createIndex(row_index, 0), row_index, row_index)
        self._fit_configurations.pop(row_index)
        self.endRemoveRows()
        self.sigFitConfigurationsChanged.emit(self.configuration_names)

    def get_configuration_by_name(self, name):
        try:
            row_index = self.configuration_names.index(name)
        except ValueError:
            raise ValueError(f'No fit configuration found with name "{name}".')
        return self._fit_configurations[row_index]

    def flags(self, index):
        if index.isValid():
            return QtCore.Qt.ItemIsEditable | QtCore.Qt.ItemIsEnabled

    def rowCount(self, parent=QtCore.QModelIndex()):
        return len(self._fit_configurations)

    def headerData(self, section, orientation, role=QtCore.Qt.DisplayRole):
        if role == QtCore.Qt.DisplayRole:
            if (orientation == QtCore.Qt.Horizontal) and (section == 0):
                return 'Fit Configurations'
            elif orientation == QtCore.Qt.Vertical:
                try:
                    return self.configuration_names[section]
                except IndexError:
                    pass
        return None

    def data(self, index=QtCore.QModelIndex(), role=QtCore.Qt.DisplayRole):
        if (role == QtCore.Qt.DisplayRole) and (index.isValid()):
            try:
                return self._fit_configurations[index.row()]
            except IndexError:
                pass
        return None

    def setData(self, index, value, role=QtCore.Qt.EditRole):
        if index.isValid():
            config = index.data(QtCore.Qt.DisplayRole)
            if config is None:
                return False
            new_params = value[1]
            params = config.default_parameters
            for name in [p for p in params if p not in new_params]:
                del params[name]
            for name, p in params.items():
                value_tuple = new_params[name]
                p.set(vary=value_tuple[0],
                      value=value_tuple[1],
                      min=value_tuple[2],
                      max=value_tuple[3])
            config.estimator = None if not value[0] else value[0]
            config.custom_parameters = None if not params else params
            self.dataChanged.emit(self.createIndex(index.row(), 0),
                                  self.createIndex(index.row(), 0))
            return True
        return False

    def dump_configs(self):
        """
        Returns all currently held fit configurations as dictionary representations containing only
        data types that can be dumped as YAML in the Qudi app status.

        Returns
        -------
        list of dict
            List of fit configuration dictionary representations.
        """
        return [cfg.to_dict() for cfg in self._fit_configurations]

    def load_configs(self, configs):
        """Initializes/overwrites all currently held fit configurations by a given iterable of dict
        representations (see also: FitConfigurationsModel.dump_configs).

        This method will reset the list model.

        Parameters
        ----------
        configs : iterable
            Iterable of FitConfiguration dictionary representations.
            See also: FitConfigurationsModel.dump_configs.

        """
        config_objects = list()
        for cfg in configs:
            try:
                config_objects.append(FitConfiguration.from_dict(cfg))
            except:
                _log.warning(f'Unable to load fit configuration:\n{cfg}')
        self.beginResetModel()
        self._fit_configurations = config_objects
        self.endResetModel()
        self.sigFitConfigurationsChanged.emit(self.configuration_names)


class FitContainer(QtCore.QObject):
    """
    """
    sigFitConfigurationsChanged = QtCore.Signal(tuple)  # config_names
    sigLastFitResultChanged = QtCore.Signal(str, object)  # (fit_config name, lmfit.ModelResult)

    def __init__(self, *args, config_model, **kwargs):
        assert isinstance(config_model, FitConfigurationsModel)
        super().__init__(*args, **kwargs)
        self._access_lock = Mutex()
        self._configuration_model = config_model
        self._last_fit_result = None
        self._last_fit_config = 'No Fit'

        self._configuration_model.sigFitConfigurationsChanged.connect(
            self.sigFitConfigurationsChanged
        )

    @property
    def fit_configurations(self):
        return self._configuration_model.configurations

    @property
    def fit_configuration_names(self):
        return self._configuration_model.configuration_names

    @property
    def last_fit(self):
        with self._access_lock:
            return self._last_fit_config, self._last_fit_result

    @QtCore.Slot(str, object, object)
    def fit_data(self, fit_config, x, data):
        with self._access_lock:
            if fit_config:
                # Handle "No Fit" case
                if fit_config == 'No Fit':
                    self._last_fit_result = None
                    self._last_fit_config = 'No Fit'
                else:
                    config = self._configuration_model.get_configuration_by_name(fit_config)
                    model = _fit_models[config.model]()
                    estimator = config.estimator
                    add_parameters = config.custom_parameters
                    if estimator is None:
                        parameters = model.make_params()
                    else:
                        parameters = model.estimators[estimator](data, x)
                    if add_parameters is not None:
                        for name, param in add_parameters.items():
                            parameters[name] = param
                    result = model.fit(data, parameters, x=x)
                    # Mutate lmfit.ModelResult object to include high-resolution result curve
                    high_res_x = np.linspace(x[0], x[-1], len(x) * 10)
                    result.high_res_best_fit = (high_res_x,
                                                model.eval(**result.best_values, x=high_res_x))
                    self._last_fit_result = result
                    self._last_fit_config = fit_config
                self.sigLastFitResultChanged.emit(self._last_fit_config, self._last_fit_result)
                return self._last_fit_config, self._last_fit_result
            return '', None

    @staticmethod
    def formatted_result(fit_result: Union[None, lmfit.model.ModelResult],
                         parameters_units: Optional[Mapping[str, str]] = None) -> str:
        if fit_result is None:
            return ''
        if parameters_units is None:
            parameters_units = dict()

        parameters_to_format = dict()
        for name, param in fit_result.params.items():
            stderr = param.stderr if param.vary else None
            stderr = np.nan if param.vary and stderr is None else stderr

            parameters_to_format[name] = {'value': param.value,
                                          'error': stderr,
                                          'unit': parameters_units.get(name, '')}

        return create_formatted_output(parameters_to_format)

    @staticmethod
    def dict_result(fit_result: Union[None, lmfit.model.ModelResult],
                    parameters_units: Optional[Mapping[str, str]] = None,
                    export_keys: Optional[Iterable[str]] = ('value', 'stderr')) -> dict:
        if fit_result is None:
            return dict()
        if parameters_units is None:
            parameters_units = dict()

        fitparams = fit_result.result.params
        export_dict = {'model': fit_result.model.name}

        for key, res in fitparams.items():
            dict_i = {key: getattr(res, key) for key in export_keys}
            dict_i['unit'] = parameters_units.get(key, '')
            export_dict[key] = dict_i

        return export_dict



# -*- coding: utf-8 -*-
"""
This file contains the Qudi Manager class.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['StatusVar', 'ConfigOption', 'Connector', 'Base', 'LogicBase', 'GuiBase', 'get_logger']

import os
from importlib import metadata
__version__ = metadata.version('qudi-core')

# Set QT_API environment variable to PySide2
os.environ['QT_API'] = 'pyside2'

from qudi.core.statusvariable import StatusVar
from qudi.core.configoption import ConfigOption
from qudi.core.connector import Connector
from qudi.core.module import Base, LogicBase, GuiBase
from qudi.core.logger import get_logger



# -*- coding: utf-8 -*-

"""
QWidget serving as editor for the default global configuration section

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['GlobalConfigWidget', 'GlobalOptionsWidget', 'RemoteServerWidget', 'CustomOptionsWidget']

from typing import Dict, Union, Mapping, Optional, Any
from PySide2 import QtCore, QtWidgets

from qudi.util.widgets.separator_lines import HorizontalLine
from qudi.util.widgets.path_line_edit import PathLineEdit
from qudi.tools.config_editor.custom_widgets import CustomOptionsWidget
from qudi.core.config.schema import config_schema


class RemoteServerWidget(QtWidgets.QWidget):
    """Remote modules server configuration editor widget.
    """
    def __init__(self,
                 config: Optional[Mapping[str, Union[int, str]]] = None,
                 parent: Optional[QtWidgets.QWidget] = None
                 ) -> None:
        super().__init__(parent=parent)

        # Create main layout
        layout = QtWidgets.QGridLayout()
        layout.setColumnStretch(1, 1)
        self.setLayout(layout)

        # server enable flag
        label = QtWidgets.QLabel('Remote modules server:')
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.enable_checkbox = QtWidgets.QCheckBox()
        self.enable_checkbox.setChecked(True)
        self.enable_checkbox.setToolTip(
            'Whether qudi should start a remote modules server at all.\nYou will not be able to '
            'communicate with remote qudi instances if this is disabled.'
        )
        self.enable_checkbox.toggled.connect(self._toggle_editors)
        layout.addWidget(label, 0, 0)
        layout.addWidget(self.enable_checkbox, 0, 1)

        # Create server config editors in their own layout
        self._server_layout = QtWidgets.QGridLayout()
        self._server_layout.setColumnStretch(1, 1)
        layout.addLayout(self._server_layout, 1, 0, 1, 2)
        label = QtWidgets.QLabel('Host address:')
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.host_lineedit = QtWidgets.QLineEdit()
        self.host_lineedit.setToolTip(
            'The host address to share qudi modules with other qudi instances'
        )
        self._server_layout.addWidget(label, 0, 0)
        self._server_layout.addWidget(self.host_lineedit, 0, 1)

        label = QtWidgets.QLabel('Port:')
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.port_spinbox = QtWidgets.QSpinBox()
        self.port_spinbox.setToolTip('Port number for the remote modules server to bind to')
        self.port_spinbox.setRange(0, 65535)
        self.port_spinbox.setValue(12345)
        self._server_layout.addWidget(label, 1, 0)
        self._server_layout.addWidget(self.port_spinbox, 1, 1)

        label = QtWidgets.QLabel('Certificate file:')
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.certfile_lineedit = PathLineEdit(dialog_caption='Select SSL Certificate File',
                                              follow_symlinks=True)
        self.certfile_lineedit.setPlaceholderText('No certificate')
        self.certfile_lineedit.setToolTip('SSL certificate file path for the remote module server')
        self._server_layout.addWidget(label, 2, 0)
        self._server_layout.addWidget(self.certfile_lineedit, 2, 1)

        label = QtWidgets.QLabel('Key file:')
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.keyfile_lineedit = PathLineEdit(dialog_caption='Select SSL Key File',
                                             follow_symlinks=True)
        self.keyfile_lineedit.setPlaceholderText('No key')
        self.keyfile_lineedit.setToolTip('SSL key file path for the remote module server')
        self._server_layout.addWidget(label, 3, 0)
        self._server_layout.addWidget(self.keyfile_lineedit, 3, 1)

        self.set_config(config)

    @property
    def config(self) -> Union[None, Dict[str, Union[int, str]]]:
        if self.enable_checkbox.isChecked():
            host = self.host_lineedit.text().strip()
            cfg = {'address': host if host else 'localhost',
                   'port'   : self.port_spinbox.value()}
            try:
                cfg['certfile'] = self.certfile_lineedit.paths[0]
            except IndexError:
                pass
            try:
                cfg['keyfile'] = self.keyfile_lineedit.paths[0]
            except IndexError:
                pass

            return cfg
        return None

    def set_config(self, config: Union[None, Mapping[str, Union[None, int, str]]]) -> None:
        if config is None:
            self.enable_checkbox.setChecked(False)
        else:
            self.enable_checkbox.setChecked(True)
            host = config.get('address', None)
            port = config.get('port', None)
            certfile = config.get('certfile', None)
            keyfile = config.get('keyfile', None)
            if host is None:
                host = 'localhost'
            if port is None:
                port = 12345
            if certfile is None or keyfile is None:
                certfile = keyfile = ''
            self.host_lineedit.setText(host)
            self.port_spinbox.setValue(port)
            self.certfile_lineedit.setText(certfile)
            self.keyfile_lineedit.setText(keyfile)

    @QtCore.Slot(bool)
    def _toggle_editors(self, enabled: bool) -> None:
        widget_iterator = (
            self._server_layout.itemAt(ii).widget() for ii in range(self._server_layout.count())
        )
        for widget in widget_iterator:
            widget.setVisible(enabled)


class GlobalOptionsWidget(QtWidgets.QWidget):
    """
    """

    def __init__(self,
                 config: Optional[Mapping[str, Any]] = None,
                 parent: Optional[QtWidgets.QWidget] = None
                 ) -> None:
        super().__init__(parent=parent)

        # Create main layout
        layout = QtWidgets.QGridLayout()
        layout.setColumnStretch(1, 1)
        self.setLayout(layout)

        # Create module server editor
        self.remote_server_editor = RemoteServerWidget()
        layout.addWidget(self.remote_server_editor, 0, 0, 1, 2)

        # Add separator
        layout.addWidget(HorizontalLine(), 1, 0, 1, 2)

        # Create local module server port editor
        label = QtWidgets.QLabel('Namespace server port:')
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.namespace_port_spinbox = QtWidgets.QSpinBox()
        self.namespace_port_spinbox.setToolTip('Port number for the local namespace server')
        self.namespace_port_spinbox.setRange(0, 65535)
        layout.addWidget(label, 2, 0)
        layout.addWidget(self.namespace_port_spinbox, 2, 1)

        # Create flag editor to enforce remote calls by value
        label = QtWidgets.QLabel('Force remote calls by value:')
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.force_calls_by_value_checkbox = QtWidgets.QCheckBox()
        self.force_calls_by_value_checkbox.setToolTip(
            'Will force all arguments from remote calls to qudi API methods to pass by value\n'
            '(serialized -> sent to qudi -> de-serialized).'
        )
        layout.addWidget(label, 3, 0)
        layout.addWidget(self.force_calls_by_value_checkbox, 3, 1)

        # Create flag editor to hide manager window upon startup
        label = QtWidgets.QLabel('Hide manager window:')
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.hide_manager_window_checkbox = QtWidgets.QCheckBox()
        self.hide_manager_window_checkbox.setToolTip(
            'Whether to suppress the qudi module manager window at startup.'
        )
        layout.addWidget(label, 4, 0)
        layout.addWidget(self.hide_manager_window_checkbox, 4, 1)

        # Create flag editor to auomatically create a data sub-directory for each day
        label = QtWidgets.QLabel('Create daily data directories:')
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.daily_data_dirs_checkbox = QtWidgets.QCheckBox()
        self.daily_data_dirs_checkbox.setToolTip(
            'Whether to automatically create daily sub-directories in the data directory for file '
            'based data storage facilities'
        )
        layout.addWidget(label, 5, 0)
        layout.addWidget(self.daily_data_dirs_checkbox, 5, 1)

        # Create default data path editor
        label = QtWidgets.QLabel('Default data directory:')
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.data_directory_lineedit = PathLineEdit(dialog_caption='Select Default Data Directory',
                                                    select_directory=True)
        self.data_directory_lineedit.setPlaceholderText('Default "<UserHome>/qudi/Data/"')
        self.data_directory_lineedit.setToolTip('Default data directory for qudi modules to save '
                                                'measurement data into.')
        layout.addWidget(label, 6, 0)
        layout.addWidget(self.data_directory_lineedit, 6, 1)

        # Create startup modules editor
        label = QtWidgets.QLabel('Startup Modules:')
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.startup_lineedit = QtWidgets.QLineEdit()
        self.startup_lineedit.setPlaceholderText('No startup modules')
        self.startup_lineedit.setToolTip('Modules to be automatically activated on qudi startup.\n'
                                         'Separate multiple module names with commas.')
        layout.addWidget(label, 7, 0)
        layout.addWidget(self.startup_lineedit, 7, 1)

        # Create stylesheet file path editor
        label = QtWidgets.QLabel('Stylesheet:')
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.stylesheet_lineedit = PathLineEdit(dialog_caption='Select QSS Stylesheet',
                                                filters='Stylesheets (*.qss)',
                                                follow_symlinks=True)
        self.stylesheet_lineedit.setPlaceholderText('Platform dependent Qt default')
        self.stylesheet_lineedit.setToolTip(
            'File path for qudi QSS stylesheet to use.\nIf just a file name is given, the file '
            'must be found in the qudi artwork resources.'
        )
        layout.addWidget(label, 8, 0)
        layout.addWidget(self.stylesheet_lineedit, 8, 1)

        # Get default config from JSON schema
        global_props = config_schema()['properties']['global']['properties']
        self._config_defaults = {
            name: prop.get('default', None) for name, prop in global_props.items()
        }
        # Fixme: Remove deprecated option manually
        del self._config_defaults['extension_paths']

        self.set_config(config)

    @property
    def config(self) -> Dict[str, Any]:
        config = {
            'remote_modules_server'      : self.remote_server_editor.config,
            'namespace_server_port'      : self.namespace_port_spinbox.value(),
            'force_remote_calls_by_value': self.force_calls_by_value_checkbox.isChecked(),
            'hide_manager_window'        : self.hide_manager_window_checkbox.isChecked(),
            'daily_data_dirs'            : self.daily_data_dirs_checkbox.isChecked(),
            'startup_modules'            : [mod.strip() for mod in
                                            self.startup_lineedit.text().split(',') if mod.strip()],
        }
        try:
            config['default_data_dir'] = self.data_directory_lineedit.paths[0]
        except IndexError:
            config['default_data_dir'] = self._config_defaults['default_data_dir']
        try:
            config['stylesheet'] = self.stylesheet_lineedit.paths[0]
        except IndexError:
            config['stylesheet'] = self._config_defaults['stylesheet']
        return config

    def set_config(self, config: Union[None, Mapping[str, Any]]):
        if config is None:
            config = dict()
        config = {
            name: config.get(name, default) for name, default in self._config_defaults.items()
        }

        self.remote_server_editor.set_config(config['remote_modules_server'])
        self.namespace_port_spinbox.setValue(config['namespace_server_port'])
        self.force_calls_by_value_checkbox.setChecked(config['force_remote_calls_by_value'])
        self.hide_manager_window_checkbox.setChecked(config['hide_manager_window'])
        self.daily_data_dirs_checkbox.setChecked(config['daily_data_dirs'])
        default_data_dir = config['default_data_dir']
        self.data_directory_lineedit.setText('' if default_data_dir is None else default_data_dir)
        stylesheet = config['stylesheet']
        self.stylesheet_lineedit.setText('' if stylesheet is None else stylesheet)
        self.startup_lineedit.setText(','.join(config['startup_modules']))


class GlobalConfigWidget(QtWidgets.QWidget):
    """
    """

    def __init__(self,
                 config: Optional[Mapping[str, Any]] = None,
                 parent: Optional[QtWidgets.QWidget] = None
                 ) -> None:
        super().__init__(parent=parent)

        # Create main layout
        layout = QtWidgets.QVBoxLayout()
        self.setLayout(layout)

        # Create Caption
        label = QtWidgets.QLabel('Global Configuration')
        label.setAlignment(QtCore.Qt.AlignCenter)
        font = label.font()
        font.setBold(True)
        font.setPointSize(font.pointSize() + 4)
        label.setFont(font)
        layout.addWidget(label)

        # Create default config editor
        self.default_options_widget = GlobalOptionsWidget()
        layout.addWidget(self.default_options_widget)

        # Remember default global config option names
        self._default_option_names = frozenset(self.default_options_widget.config)

        # Add separator
        layout.addWidget(HorizontalLine())

        # Create custom option editor
        self.custom_options_widget = CustomOptionsWidget(forbidden_names=self._default_option_names)
        layout.addWidget(self.custom_options_widget)

        layout.addStretch(1)

        self.set_config(config)

    @property
    def config(self) -> Dict[str, Any]:
        config = self.default_options_widget.config
        config.update(self.custom_options_widget.config)
        return config

    def set_config(self, config: Union[None, Mapping[str, Any]]) -> None:
        if config is None:
            self.default_options_widget.set_config(None)
            self.custom_options_widget.set_config(None)
        else:
            default_config = {name: value for name, value in config.items() if
                              name in self._default_option_names}
            custom_config = {name: value for name, value in config.items() if
                             name not in self._default_option_names}
            self.default_options_widget.set_config(default_config)
            self.custom_options_widget.set_config(custom_config)
